# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.linalg import Vectors
from pyspark.ml.feature import VectorAssembler
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlVectorAssembler(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_vector_assembler(self):
        col_names = ["a", "b"]
        model = VectorAssembler(inputCols=col_names, outputCol="features")
        data = self.spark.createDataFrame(
            [(1.0, Vectors.dense([1.0, 2.0, 3.0]))], col_names
        )
        model_onnx = convert_sparkml(
            model,
            "Sparkml VectorAssembler",
            [("a", FloatTensorType([None, 1])), ("b", FloatTensorType([None, 3]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)
        # run the model
        predicted = model.transform(data)
        expected = (
            predicted.select("features")
            .toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values
        )
        data_np = {
            "a": data.select("a").toPandas().values.astype(numpy.float32),
            "b": data.select("b")
            .toPandas()["b"]
            .apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32),
        }
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlVectorAssembler"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["features"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)
        assert output_shapes == [[None, 4]]


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import copy

from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator
from ...common._topology import Operator, Scope
from pyspark.ml.feature import ImputerModel


def convert_imputer(scope: Scope, operator: Operator, container):
    op: ImputerModel = operator.raw_operator
    op_type = "Imputer"
    name = scope.get_unique_operator_name(op_type)
    attrs = {"name": name}
    input_type = operator.inputs[0].type
    surrogates = op.surrogateDF.toPandas().values[0].tolist()
    value = op.getOrDefault("missingValue")

    if isinstance(input_type, FloatTensorType):
        attrs["imputed_value_floats"] = surrogates
        attrs["replaced_value_float"] = value
    elif isinstance(input_type, Int64TensorType):
        attrs["imputed_value_int64s"] = [int(x) for x in surrogates]
        attrs["replaced_value_int64"] = int(value)
    else:
        raise RuntimeError("Invalid input type: " + input_type)

    if len(operator.inputs) > 1:
        concatenated_output = scope.get_unique_variable_name("concat_tensor")
        container.add_node(
            "Concat",
            operator.input_full_names,
            concatenated_output,
            name=scope.get_unique_operator_name("Concat"),
            op_version=4,
            axis=1,
        )
        imputed_output = scope.get_unique_variable_name("imputed_tensor")
        container.add_node(
            op_type,
            concatenated_output,
            imputed_output,
            op_domain="ai.onnx.ml",
            **attrs
        )
        container.add_node(
            "Split",
            imputed_output,
            operator.output_full_names,
            name=scope.get_unique_operator_name("Split"),
            op_version=2,
            axis=1,
            split=[1] * len(operator.output_full_names),
        )
    else:
        container.add_node(
            op_type,
            operator.inputs[0].full_name,
            operator.output_full_names[0],
            op_domain="ai.onnx.ml",
            **attrs
        )


register_converter("pyspark.ml.feature.ImputerModel", convert_imputer)


def calculate_imputer_output_shapes(operator):
    check_input_and_output_numbers(
        operator, output_count_range=[1, len(operator.outputs)]
    )
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, Int64TensorType],
        good_output_types=[FloatTensorType, Int64TensorType],
    )
    input_type = copy.deepcopy(operator.inputs[0].type)
    for output in operator.outputs:
        output.type = input_type


register_shape_calculator(
    "pyspark.ml.feature.ImputerModel", calculate_imputer_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import logging
from ...common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ...common._registration import register_converter, register_shape_calculator
from .decision_tree_classifier import save_read_sparkml_model_data
from .decision_tree_regressor import calculate_decision_tree_regressor_output_shapes
from .tree_ensemble_common import sparkml_tree_dataset_to_sklearn
from .tree_helper import rewrite_ids_and_process

logger = logging.getLogger("onnxmltools")


def convert_random_forest_regressor(scope, operator, container):
    op = operator.raw_operator
    op_type = "TreeEnsembleRegressor"

    main_attrs = get_default_tree_regressor_attribute_pairs()
    main_attrs["name"] = scope.get_unique_operator_name(op_type)
    main_attrs["n_targets"] = 1

    # random forest calculate the final score by averaging over all trees'
    # outcomes, so all trees' weights are identical.
    tree_weight = 1.0 / op.getNumTrees

    for tree_id in range(0, op.getNumTrees):
        tree_model = op.trees[tree_id]
        tree_df = save_read_sparkml_model_data(
            operator.raw_params["SparkSession"], tree_model
        )
        tree = sparkml_tree_dataset_to_sklearn(tree_df, is_classifier=False)
        attrs = get_default_tree_regressor_attribute_pairs()
        attrs["name"] = scope.get_unique_operator_name(op_type)
        attrs["n_targets"] = 1
        add_tree_to_attribute_pairs(attrs, False, tree, tree_id, tree_weight, 0, False)
        new_attrs = rewrite_ids_and_process(attrs, logger)
        for k, v in new_attrs.items():
            if isinstance(v, list):
                main_attrs[k].extend(v)

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names[0],
        op_domain="ai.onnx.ml",
        **main_attrs
    )


register_converter(
    "pyspark.ml.regression.RandomForestRegressionModel", convert_random_forest_regressor
)


register_shape_calculator(
    "pyspark.ml.regression.RandomForestRegressionModel",
    calculate_decision_tree_regressor_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import copy
from ...common._registration import register_shape_calculator
from ...common.utils import check_input_and_output_numbers


def calculate_identity_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)

    input = operator.inputs[0]
    output = operator.outputs[0]

    doc_string = output.type.doc_string
    output.type = copy.deepcopy(input.type)
    output.type.doc_string = doc_string


register_shape_calculator("identity", calculate_identity_output_shapes)
register_shape_calculator("imputer", calculate_identity_output_shapes)
register_shape_calculator("scaler", calculate_identity_output_shapes)
register_shape_calculator("normalizer", calculate_identity_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import collections

from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_matmul, apply_exp, apply_add
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers


def convert_aft_survival_regression(scope, operator, container):
    op = operator.raw_operator

    coefficients = op.coefficients.toArray().astype(float)
    coefficients_tensor = scope.get_unique_variable_name("coefficients_tensor")
    container.add_initializer(
        coefficients_tensor,
        onnx_proto.TensorProto.FLOAT,
        [1, len(coefficients)],
        coefficients,
    )
    intercepts = (
        op.intercept.astype(float)
        if isinstance(op.intercept, collections.abc.Iterable)
        else [float(op.intercept)]
    )
    intercepts_tensor = scope.get_unique_variable_name("intercepts_tensor")
    container.add_initializer(
        intercepts_tensor, onnx_proto.TensorProto.FLOAT, [len(intercepts)], intercepts
    )

    matmul_result = scope.get_unique_variable_name("matmul_result_tensor")
    apply_matmul(
        scope,
        [operator.input_full_names[0], coefficients_tensor],
        matmul_result,
        container,
    )
    add_result = scope.get_unique_variable_name("intercept_added_tensor")
    apply_add(scope, [matmul_result, intercepts_tensor], add_result, container)
    apply_exp(scope, add_result, operator.output_full_names, container)


register_converter(
    "pyspark.ml.regression.AFTSurvivalRegressionModel", convert_aft_survival_regression
)


def calculate_aft_survival_regression_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, 1])


register_shape_calculator(
    "pyspark.ml.regression.AFTSurvivalRegressionModel",
    calculate_aft_survival_regression_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_mul
from ....common._registration import register_converter


def convert_dot(scope, operator, container):
    # To calculate cosine similarity, we first use
    # LpNormalization to make the two input vectors unit-length.
    # Then, we calculate element-wise product of the two
    # unit-length vectors. Finally, the similarity is the
    # sum of all the product's elements. Notice that
    # we carefully specify the axis of the subsequent operators,
    # so they can work properly with a batch of vectors.

    if operator.raw_operator.dot.cosineSimilarity:
        # Normalize the first input and store the result on a temporal variable
        intra_variable_name1 = scope.get_unique_variable_name(
            operator.inputs[0].full_name + "_normalized"
        )
        normalizer_name1 = scope.get_unique_operator_name("L2NormNormalizer")
        attrs1 = {"name": normalizer_name1, "p": 2, "aixs": 1}
        container.add_node(
            "LpNormalization",
            [operator.inputs[0].full_name],
            [intra_variable_name1],
            **attrs1
        )

        # Normalize the second input and store the result on a temporal variable
        intra_variable_name2 = scope.get_unique_variable_name(
            operator.inputs[1].full_name + "_normalized"
        )
        normalizer_name2 = scope.get_unique_operator_name("L2NormNormalizer")
        attrs2 = {"name": normalizer_name2, "p": 2, "aixs": 1}
        container.add_node(
            "LpNormalization",
            [operator.inputs[1].full_name],
            [intra_variable_name2],
            **attrs2
        )
    else:
        # This case is a simple dot product; no normalization is required.
        intra_variable_name1 = operator.inputs[0].full_name
        intra_variable_name2 = operator.inputs[1].full_name

    # Do element-wise product of the two unit-length tensors
    product_name = scope.get_unique_variable_name(
        intra_variable_name1 + "_multiply_" + intra_variable_name2
    )
    apply_mul(
        scope,
        [intra_variable_name1, intra_variable_name2],
        product_name,
        container,
        broadcast=0,
    )

    # Sum up results from different dimensions to get the final cosine similarity
    reducer_name = scope.get_unique_operator_name("ReduceSum")
    reducer_attrs = {"name": reducer_name, "axes": [1], "keepdims": 1}
    container.add_node(
        "ReduceSum", [product_name], operator.output_full_names, **reducer_attrs
    )


register_converter("dot", convert_dot)



# SPDX-License-Identifier: Apache-2.0

from onnxconverter_common.container import (
    RawModelContainer,
    CommonSklearnModelContainer,
)


class LightGbmModelContainer(CommonSklearnModelContainer):
    pass


class XGBoostModelContainer(CommonSklearnModelContainer):
    pass


class H2OModelContainer(CommonSklearnModelContainer):
    pass


class SparkmlModelContainer(RawModelContainer):
    def __init__(self, sparkml_model):
        super(SparkmlModelContainer, self).__init__(sparkml_model)
        # Sparkml models have no input and output specified,
        # so we create them and store them in this container.
        self._inputs = []
        self._outputs = []

    @property
    def input_names(self):
        return [variable.raw_name for variable in self._inputs]

    @property
    def output_names(self):
        return [variable.raw_name for variable in self._outputs]

    def add_input(self, variable):
        # The order of adding variables matters. The final model's
        # input names are sequentially added as this list
        if variable not in self._inputs:
            self._inputs.append(variable)

    def add_output(self, variable):
        # The order of adding variables matters.
        # The final model's output names are sequentially added as this list
        if variable not in self._outputs:
            self._outputs.append(variable)


class CoremlModelContainer(RawModelContainer):
    def __init__(self, coreml_model):
        super(CoremlModelContainer, self).__init__(coreml_model)

    @property
    def input_names(self):
        return [str(var.name) for var in self.raw_model.description.input]

    @property
    def output_names(self):
        return [str(var.name) for var in self.raw_model.description.output]


class LibSvmModelContainer(CommonSklearnModelContainer):
    pass



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import inspect
import os
import numpy
import pandas
from pyspark.ml.classification import (
    MultilayerPerceptronClassifier,
    MultilayerPerceptronClassificationModel,
)
from pyspark.ml.linalg import VectorUDT, SparseVector
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlMLPClassifier(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_mlp_classifier_binary_class(self):
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)
        #
        # truncate the features
        #
        self.spark.udf.register(
            "truncateFeatures",
            lambda x: SparseVector(100, range(0, 100), x.toArray()[30:130]),
            VectorUDT(),
        )

        data = original_data.selectExpr(
            "label", "truncateFeatures(features) as features"
        )

        mlp = MultilayerPerceptronClassifier(
            maxIter=100,
            tol=0.0001,
            seed=137,
            layers=[100, 20, 5, 2],
        )
        model: MultilayerPerceptronClassificationModel = mlp.fit(data)

        # the name of the input for Logistic Regression is 'features'
        C = model.numFeatures
        model_onnx = convert_sparkml(
            model,
            "sparkml multilayer perceptron classifier",
            [("features", FloatTensorType([None, C]))],
            target_opset=TARGET_OPSET,
        )

        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        # predicted.select("prediction", "probability", "label").show(
        #       100, truncate=False)

        data_np = (
            data.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas()
            .probability.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32),
        ]

        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlMLPClassifier"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(
            ["prediction", "probability"], data_np, onnx_model_path
        )
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.feature import CountVectorizer, CountVectorizerModel
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import StringTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase

TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlCountVectorizer(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_count_vectorizer_default(self):
        data = self.spark.createDataFrame(
            [
                ("A B C".split(" "),),
                ("A B B C A".split(" "),),
            ],
            ["text"],
        )
        count_vec = CountVectorizer(
            inputCol="text", outputCol="result", minTF=1.0, binary=False
        )
        model: CountVectorizerModel = count_vec.fit(data)
        result = model.transform(data)

        model_onnx = convert_sparkml(
            model,
            "Sparkml CountVectorizer",
            [("text", StringTensorType([None, None]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        data_pd = data.toPandas()
        data_np = {
            "text": data_pd.text.apply(lambda x: pandas.Series(x)).values.astype(str),
        }

        expected = {
            "prediction_result": numpy.asarray(
                result.toPandas()
                .result.apply(lambda x: pandas.Series(x.toArray()))
                .values.astype(numpy.float32)
            ),
        }

        paths = save_data_models(
            data_np,
            expected,
            model,
            model_onnx,
            basename="SparkmlCountVectorizerModel_Default",
        )
        onnx_model_path = paths[-1]

        output_names = ["result"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))

        assert output_shapes[0] == [None, 3]
        compare_results(
            expected["prediction_result"], actual_output["result"], decimal=5
        )

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_count_vectorizer_binary(self):
        data = self.spark.createDataFrame(
            [
                ("A B C".split(" "),),
                ("A B B C A".split(" "),),
                ("B B B D".split(" "),),
            ],
            ["text"],
        )
        count_vec = CountVectorizer(
            inputCol="text", outputCol="result", minTF=2.0, binary=True
        )
        model: CountVectorizerModel = count_vec.fit(data)
        result = model.transform(data)

        model_onnx = convert_sparkml(
            model,
            "Sparkml CountVectorizer",
            [("text", StringTensorType([None, None]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        data_pd = data.toPandas()
        data_np = {
            "text": data_pd.text.apply(lambda x: pandas.Series(x)).values.astype(str),
        }

        expected = {
            "prediction_result": numpy.asarray(
                result.toPandas()
                .result.apply(lambda x: pandas.Series(x.toArray()))
                .values.astype(numpy.float32)
            ),
        }

        paths = save_data_models(
            data_np,
            expected,
            model,
            model_onnx,
            basename="SparkmlCountVectorizerModel_Binary",
        )
        onnx_model_path = paths[-1]

        output_names = ["result"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))

        assert output_shapes[0] == [None, 4]
        compare_results(
            expected["prediction_result"], actual_output["result"], decimal=5
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_batch_norm, apply_instance_norm
from ....common._registration import register_converter


def convert_batch_normalization(scope, operator, container):
    params = operator.raw_operator.batchnorm

    if params.instanceNormalization and not params.computeMeanVar:
        raise ValueError(
            "It is impossible to do instance normalization "
            "without re-computing mean and variance"
        )

    if params.instanceNormalization and params.computeMeanVar:
        op_type = "InstanceNormalization"
    else:
        op_type = "BatchNormalization"

    inputs = [operator.inputs[0].full_name]
    outputs = [operator.outputs[0].full_name]
    scale_tensor_name = scope.get_unique_variable_name(op_type + "_scale")
    container.add_initializer(
        scale_tensor_name,
        onnx_proto.TensorProto.FLOAT,
        [params.channels],
        params.gamma.floatValue,
    )
    inputs.append(scale_tensor_name)
    bias_tensor_name = scope.get_unique_variable_name(op_type + "_B")
    container.add_initializer(
        bias_tensor_name,
        onnx_proto.TensorProto.FLOAT,
        [params.channels],
        params.beta.floatValue,
    )
    inputs.append(bias_tensor_name)

    epsilon = params.epsilon

    if op_type == "BatchNormalization":
        mean_tensor_name = scope.get_unique_variable_name(op_type + "_mean")
        container.add_initializer(
            mean_tensor_name,
            onnx_proto.TensorProto.FLOAT,
            [params.channels],
            params.mean.floatValue,
        )
        inputs.append(mean_tensor_name)
        variance_tensor_name = scope.get_unique_variable_name(op_type + "_variance")
        container.add_initializer(
            variance_tensor_name,
            onnx_proto.TensorProto.FLOAT,
            [params.channels],
            params.variance.floatValue,
        )
        inputs.append(variance_tensor_name)
        momentum = 0.0
        spatial = 1  # True

        if not params.instanceNormalization and params.computeMeanVar:
            # In this case, we apply batch normalization and adjust
            # the statistics stored according the the batch
            # being processed.

            # To update "mean" and "var," we put their updated results
            # back to the associated input tensors.
            outputs += inputs[1:3]
            # We also allocate two extra output buffers to store some
            # intermediate results, but they are not used
            # in CoreML model.
            outputs.append(scope.get_unique_variable_name("saved_mean"))
            outputs.append(scope.get_unique_variable_name("saved_var"))
            # We choose "training" mode because some variables need to be updated.
            is_test = 0  # False
        elif not params.instanceNormalization and not params.computeMeanVar:
            # In this case, batch normalization is applied without
            # updating mean, variance, etc. according to
            # the batches being processed. It means this operator
            # works under testing model. Because there is no
            # variable update, we don't need to specify extra
            # inputs and outputs like in previous code block.
            is_test = 1  # True
        else:
            raise ValueError("Unsupported operation mode")

        apply_batch_norm(
            scope,
            inputs,
            outputs,
            container,
            operator_name=operator.full_name,
            epsilon=epsilon,
            is_test=is_test,
            momentum=momentum,
            spatial=spatial,
        )
    else:
        apply_instance_norm(
            scope,
            inputs,
            outputs,
            container,
            operator_name=operator.full_name,
            epsilon=epsilon,
        )


register_converter("batchnorm", convert_batch_normalization)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_concat
from ....common._registration import register_converter


def convert_concat(scope, operator, container):
    if operator.raw_operator.concat.sequenceConcat:
        axis = 0
    else:
        axis = 1

    apply_concat(
        scope,
        operator.input_full_names,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        axis=axis,
    )


register_converter("concat", convert_concat)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_identical_float_tensor_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input = operator.inputs[0]
    output = operator.outputs[0]

    doc_string = output.type.doc_string
    output.type.shape = copy.deepcopy(
        input.type.shape
    )  # Similar to identity but only accept floats
    output.type.doc_string = doc_string


# Preprocessing layers in CoreML
register_shape_calculator("scalerPreprocessor", calculate_identical_float_tensor_shapes)
register_shape_calculator(
    "meanImagePreprocessor", calculate_identical_float_tensor_shapes
)

# Standard neural network layers
register_shape_calculator("activation", calculate_identical_float_tensor_shapes)
register_shape_calculator("bias", calculate_identical_float_tensor_shapes)
register_shape_calculator("l2normalize", calculate_identical_float_tensor_shapes)
register_shape_calculator("lrn", calculate_identical_float_tensor_shapes)
register_shape_calculator("mvn", calculate_identical_float_tensor_shapes)
register_shape_calculator("scale", calculate_identical_float_tensor_shapes)
register_shape_calculator("softmax", calculate_identical_float_tensor_shapes)
register_shape_calculator("unary", calculate_identical_float_tensor_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_normalizer(scope, operator, container):
    op_type = "Normalizer"
    attrs = {"name": operator.full_name}
    norms = ["MAX", "L1", "L2"]
    norm_type = operator.raw_operator.normalizer.normType
    if norm_type in range(3):
        attrs["norm"] = norms[norm_type]
    else:
        raise ValueError("Invalid norm type: " + norm_type)

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("normalizer", convert_normalizer)



# SPDX-License-Identifier: Apache-2.0

"""
Helpers to help testing deep learning converted models.
"""
import numpy as np
from .main import save_model

rt_onnxruntime, rt_caffe2, rt_cntk = range(3)
_rt_installed = None


def find_inference_engine():
    """
    Finds an available backend to evaluate deep learning models.
    Returns its name as a string.
    """
    global _rt_installed
    if _rt_installed is not None:
        return _rt_installed

    try:
        pass

        _rt_installed = rt_onnxruntime
    except ImportError:
        try:
            pass

            _rt_installed = rt_cntk
        except ImportError:
            try:
                pass

                _rt_installed = rt_caffe2
            except ImportError:
                pass

    return _rt_installed


def evaluate_deep_model(onnx_model, inputs, rt_type=None):
    """
    Evaluates a deep learning model with
    onnxruntime, *CNTK* or *caffe2*
    """
    if rt_type is None:
        rt_type = find_inference_engine()
    if rt_type == rt_onnxruntime:
        return _evaluate_onnxruntime(onnx_model, inputs)
    if rt_type == rt_cntk:
        return _evaluate_cntk(onnx_model, inputs)
    elif rt_type == rt_caffe2:
        return _evaluate_caffe2(onnx_model, inputs)
    else:
        raise ImportError("No runtime found. Need either CNTK or Caffe2")


def _evaluate_onnxruntime(onnx_model, inputs):
    import onnxruntime

    runtime = onnxruntime.InferenceSession(
        onnx_model.SerializeToString(), providers=["CPUExecutionProvider"]
    )
    result = None
    inputs = inputs if isinstance(inputs, list) else [inputs]
    for i_ in range(
        inputs[0].shape[0]
    ):  # TODO: onnxruntime can't support batch_size > 1
        out = runtime.run(
            [],
            {
                x.name: inputs[n_][i_ : i_ + 1]
                for n_, x in enumerate(runtime.get_inputs())
            },
        )[0]
        result = out if result is None else np.concatenate((result, out))

    return result[0] if isinstance(result, list) else result


def _evaluate_caffe2(onnx_model, inputs):
    from caffe2.python.onnx.backend import run_model

    if not isinstance(inputs, list):
        inputs = [inputs]

    outputs = run_model(onnx_model, inputs)

    adjusted_outputs = dict()
    for output in onnx_model.graph.output:
        adjusted_outputs[output.name] = outputs[output.name]

    return adjusted_outputs[onnx_model.graph.output[0].name]


def _evaluate_cntk(onnx_model, inputs):
    import cntk

    if not isinstance(inputs, list):
        inputs = [inputs]

    adjusted_inputs = dict()

    for i, x in enumerate(inputs):
        onnx_name = onnx_model.graph.input[i].name
        adjusted_inputs[onnx_name] = [
            np.ascontiguousarray(np.squeeze(_, axis=0)) for _ in np.split(x, x.shape[0])
        ]

    temporary_onnx_model_file_name = "temp_" + onnx_model.graph.name + ".onnx"
    save_model(onnx_model, temporary_onnx_model_file_name)
    cntk_model = cntk.Function.load(
        temporary_onnx_model_file_name, format=cntk.ModelFormat.ONNX
    )

    return cntk_model.eval(adjusted_inputs)


def create_tensor(N, C, H=None, W=None):
    if H is None and W is None:
        return np.random.rand(N, C).astype(np.float32, copy=False)
    elif H is not None and W is not None:
        return np.random.rand(N, C, H, W).astype(np.float32, copy=False)
    else:
        raise ValueError("This function only produce 2-D or 4-D tensor")



# SPDX-License-Identifier: Apache-2.0

import logging
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator
from .tree_ensemble_common import (
    save_read_sparkml_model_data,
    sparkml_tree_dataset_to_sklearn,
    add_tree_to_attribute_pairs,
    get_default_tree_classifier_attribute_pairs,
)
from .tree_helper import rewrite_ids_and_process

logger = logging.getLogger("onnxmltools")


def convert_decision_tree_classifier(scope, operator, container):
    op = operator.raw_operator
    op_type = "TreeEnsembleClassifier"

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)
    attrs["classlabels_int64s"] = list(range(0, op.numClasses))

    logger.info("[convert_decision_tree_classifier] save_read_sparkml_model_data")
    tree_df = save_read_sparkml_model_data(operator.raw_params["SparkSession"], op)
    logger.info("[convert_decision_tree_classifier] sparkml_tree_dataset_to_sklearn")
    tree = sparkml_tree_dataset_to_sklearn(tree_df, is_classifier=True)
    logger.info("[convert_decision_tree_classifier] add_tree_to_attribute_pairs")
    add_tree_to_attribute_pairs(
        attrs, True, tree, 0, 1.0, 0, leaf_weights_are_counts=True
    )
    logger.info(
        "[convert_decision_tree_classifier] n_nodes=%d", len(attrs["nodes_nodeids"])
    )

    # Some values appear in an array of one element instead of a float.

    new_attrs = rewrite_ids_and_process(attrs, logger)

    container.add_node(
        op_type,
        operator.input_full_names,
        [operator.outputs[0].full_name, operator.outputs[1].full_name],
        op_domain="ai.onnx.ml",
        **new_attrs
    )


register_converter(
    "pyspark.ml.classification.DecisionTreeClassificationModel",
    convert_decision_tree_classifier,
)


def calculate_decision_tree_classifier_output_shapes(operator):
    check_input_and_output_numbers(
        operator, input_count_range=1, output_count_range=[1, 2]
    )
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]

    class_count = operator.raw_operator.numClasses
    operator.outputs[0].type = Int64TensorType(shape=[N])
    operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator(
    "pyspark.ml.classification.DecisionTreeClassificationModel",
    calculate_decision_tree_classifier_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_dictionary_vectorizer(scope, operator, container):
    op_type = "DictVectorizer"
    attrs = {"name": operator.full_name}
    raw_model = operator.raw_operator.dictVectorizer
    if raw_model.HasField("stringToIndex"):
        attrs["string_vocabulary"] = raw_model.stringToIndex.vector
    else:
        attrs["int64_vocabulary"] = raw_model.int64ToIndex.vector

    container.add_node(
        op_type,
        [operator.inputs[0].full_name],
        [operator.outputs[0].full_name],
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("dictVectorizer", convert_dictionary_vectorizer)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_concat, apply_argmax
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from .linear_classifier import convert_sparkml_linear_classifier
from ..ops_names import get_sparkml_operator_name
from ...common._registration import register_converter, register_shape_calculator


def convert_one_vs_rest(scope, operator, container):
    op = operator.raw_operator
    classifier_output_names = []
    # initializer needed to extract the 2nd value of probability array
    index_tensor = scope.get_unique_variable_name("index_tensor")
    container.add_initializer(index_tensor, onnx_proto.TensorProto.INT64, [1], [1])
    # OneVsRest could have different classifiers
    # all must have at least Probability values for us to do the argmax
    for sub_model in op.models:
        classifier_op = scope.declare_local_operator(
            get_sparkml_operator_name(type(sub_model)), sub_model
        )
        classifier_op.raw_params = operator.raw_params
        classifier_op.inputs = operator.inputs
        classifier_prediction_output = scope.declare_local_variable(
            "classifier_prediction", FloatTensorType()
        )
        classifier_probability_output = scope.declare_local_variable(
            "classifier_probability", FloatTensorType()
        )
        classifier_op.outputs.append(classifier_prediction_output)
        classifier_op.outputs.append(classifier_probability_output)
        convert_sparkml_linear_classifier(scope, classifier_op, container)
        classifier_op.is_evaluated = True
        single_feature_tensor = scope.get_unique_variable_name("single_feature_tensor")
        container.add_node(
            "ArrayFeatureExtractor",
            [classifier_probability_output.full_name, index_tensor],
            single_feature_tensor,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )
        classifier_output_names.append(single_feature_tensor)

    concatenated_probabilities = scope.get_unique_variable_name(
        "concatenated_predictions_tensor"
    )
    apply_concat(
        scope, classifier_output_names, concatenated_probabilities, container, axis=1
    )
    # to get Prediction from probability
    apply_argmax(
        scope,
        concatenated_probabilities,
        operator.outputs[0].full_name,
        container,
        axis=1,
        keepdims=1,
    )


register_converter("pyspark.ml.classification.OneVsRestModel", convert_one_vs_rest)


def calculate_one_vs_rest_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])


register_shape_calculator(
    "pyspark.ml.classification.OneVsRestModel", calculate_one_vs_rest_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import pandas
import numpy
from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_add, apply_mul, apply_sum
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import StringTensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def convert_word2vec(scope, operator, container):
    op = operator.raw_operator
    vectors = (
        op.getVectors()
        .toPandas()
        .vector.apply(lambda x: pandas.Series(x.toArray()))
        .values.astype(numpy.float32)
    )
    cats_strings = op.getVectors().toPandas().word.values.tolist()
    cats_int64s = [x for x in range(0, len(cats_strings))]
    word_count = operator.inputs[0].type.shape[1]

    vectors_tensor = scope.get_unique_variable_name("vectors_tensor")
    container.add_initializer(
        vectors_tensor, onnx_proto.TensorProto.FLOAT, vectors.shape, vectors.flatten()
    )
    word_indices = scope.get_unique_variable_name("word_indices_tensor")
    container.add_node(
        "CategoryMapper",
        operator.input_full_names,
        word_indices,
        op_domain="ai.onnx.ml",
        cats_int64s=cats_int64s,
        cats_strings=cats_strings,
        default_int64=-1,
    )
    one = scope.get_unique_variable_name("one_tensor")
    container.add_initializer(one, onnx_proto.TensorProto.INT64, [1], [1])
    zero = scope.get_unique_variable_name("zero_tensor")
    container.add_initializer(zero, onnx_proto.TensorProto.INT64, [1], [0])

    sliced_outputs = []
    for i in range(0, word_count):
        index = scope.get_unique_variable_name("index_tensor")
        container.add_initializer(index, onnx_proto.TensorProto.INT64, [1], [i])
        selected_index = scope.get_unique_variable_name("selected_index_tensor")
        container.add_node(
            "ArrayFeatureExtractor",
            [word_indices, index],
            selected_index,
            op_domain="ai.onnx.ml",
        )
        reshaped_index = scope.get_unique_variable_name("reshaped_tensor")
        container.add_node(
            "Reshape", [selected_index, one], reshaped_index, op_version=5
        )
        end_index = scope.get_unique_variable_name("end_index_tensor")
        apply_add(scope, [one, reshaped_index], end_index, container, axis=0)
        sliced_output = scope.get_unique_variable_name("sliced_tensor")
        container.add_node(
            "DynamicSlice",
            [vectors_tensor, reshaped_index, end_index, zero],
            sliced_output,
        )
        sliced_outputs.append(sliced_output)

    sum_vector = scope.get_unique_variable_name("sum_tensor")
    apply_sum(scope, sliced_outputs, sum_vector, container)

    factor = scope.get_unique_variable_name("factor_tensor")
    container.add_initializer(
        factor,
        onnx_proto.TensorProto.FLOAT,
        [1],
        [1 / operator.inputs[0].type.shape[1]],
    )
    apply_mul(scope, [factor, sum_vector], operator.output_full_names, container)


register_converter("pyspark.ml.feature.Word2VecModel", convert_word2vec)


def calculate_word2vec_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[StringTensorType])

    N = operator.inputs[0].type.shape[0]
    C = operator.raw_operator.getOrDefault("vectorSize")
    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator(
    "pyspark.ml.feature.Word2VecModel", calculate_word2vec_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_add, apply_mul
from ....common._registration import register_converter


def deduce_broadcast_axis_and_shape(target_opset, shape):
    # This function is used to calculate the first axis aligned
    # with the scalar and the scalar's ONNX shape for reduce-
    # like operators. Assuming input variable is always a 4-D tensor,
    # we provide a few of examples. If scalar's shape
    # is [1, 2, 3] and input shape is [5, 2, 3, 8], the aligned axis
    # is the [2] (indexed by 1 because indexes are 0-based)
    # in [5, 2, 3, 8], and the desired scalar shape in ONNX is [2, 3]
    # (the [1] in [1, 2, 3] is redundant and can cause
    # errors in ONNX's boardcasting). If the scaler's shape is [1],
    # no matter what shape the input is, we leave the axis
    # "None" because ONNX operator may automatically handle it.
    # After ONNX-1.2, we adopt Numpy-style broadcasting rule.

    if target_opset < 7:
        # Input shape is [N, C, H, W]
        if len(shape) == 1:
            if shape[0] == 1:
                # shape is [] for scalar, we don't specify axis because it's a scalar
                return None, []
            else:
                # shape is [C], alignment starting at C-axis (indexed by 1)
                return 1, shape
        elif len(shape) == 3:
            if shape[0] == 1:
                # shape is [1, H, W], alignment starting at H-axis (indexed by 2)
                return 2, [shape[1], shape[2]]
            else:
                # shape is [C, H, W], alignment starting at C-axis (indexed by 1)
                return 1, shape
    else:
        # Input shape is [N, C, H, W]
        if len(shape) == 1:
            if shape[0] == 1:
                # shape is [1] for scalar, we don't specify axis because it's a scalar
                return None, [1]
            else:
                # shape is [C], alignment starting at C-axis
                return None, [shape[0], 1, 1]
        elif len(shape) == 3:
            if shape[0] == 1:
                # shape is [1, H, W], alignment starting at C-axis
                return None, shape
            else:
                # shape is [C, H, W], alignment starting at C-axis
                return None, shape


def convert_scale(scope, operator, container):
    # In CoreML's ScaleLayer, the input is first scaled by their
    # "scale" attribute and then a "bias" can be added.
    # Symbols:
    #  a: scale attribute in CoreML's ScaleLayer
    #  b: bias attribute in CoreML's ScaleLayer
    #  x: input
    #  y: output
    # The math formulation of ScaleLayer should be
    #  y = a * x + b
    # Therefore, our strategy of composing ScaleLayer is to
    # have one multiplication followed by an addition.
    params = operator.raw_operator.scale
    op1_type = "Mul"
    scale_axis, scale_shape = deduce_broadcast_axis_and_shape(
        container.target_opset, params.shapeScale
    )
    scale_name = scope.get_unique_variable_name(op1_type + "_B")
    container.add_initializer(
        scale_name, onnx_proto.TensorProto.FLOAT, scale_shape, params.scale.floatValue
    )

    # CoreML is at most 3-D, so we always turn broadcasting on.
    scale_broadcast = 1

    if not params.hasBias:
        # Create a element-wise multiplication and use it to scale
        # the input. The first input is the variable we want
        # to scale while the second input is their multipliers.
        apply_mul(
            scope,
            [operator.inputs[0].full_name, scale_name],
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
            axis=scale_axis,
            broadcast=scale_broadcast,
        )
    else:
        # Declare a temporal variable to store the scaled input
        intra_variable_name = scope.get_unique_variable_name(
            operator.inputs[0].full_name + "_scaled"
        )
        # Create a element-wise multiplication and use it to
        # scale the input and save the result to a temporal variable
        apply_mul(
            scope,
            [operator.inputs[0].full_name, scale_name],
            intra_variable_name,
            container,
            operator_name=operator.full_name,
            axis=scale_axis,
            broadcast=scale_broadcast,
        )

        # Prepare materials to build an Add operator for adding bias
        bias_axis, bias_shape = deduce_broadcast_axis_and_shape(
            container.target_opset, params.shapeBias
        )

        # CoreML is at most 3-D, so we always turn broadcasting on.
        bias_broadcast = 1

        bias_name = scope.get_unique_variable_name(operator.full_name + "_B")
        container.add_initializer(
            bias_name, onnx_proto.TensorProto.FLOAT, bias_shape, params.bias.floatValue
        )
        # As bias exists, we add the bias into the output of
        # the multiplication and then use the output of addition
        # as the final output of this conversion.
        apply_add(
            scope,
            [intra_variable_name, bias_name],
            operator.output_full_names,
            container,
            axis=bias_axis,
            broadcast=bias_broadcast,
        )


register_converter("scale", convert_scale)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_sub
from ....common._registration import register_converter


def convert_preprocessing_mean_image(scope, operator, container):
    mean_tensor_name = scope.get_unique_variable_name(operator.full_name + "_mean")

    # We assume that the first input's shape is [N, C, H, W]
    # so that the mean image's shape, [C, H, W], can
    # be inferred from the first input's shape.
    container.add_initializer(
        mean_tensor_name,
        onnx_proto.TensorProto.FLOAT,
        operator.inputs[0].type.shape[1:],
        operator.raw_operator.meanImage,
    )

    # We assume that the first input variable's shape is [N, C, H, W]
    # while the mean image's shape is [C, H, W]. Thus,
    # broadcasting should be enabled starting with axis=1.
    apply_sub(
        scope,
        [operator.inputs[0].full_name, mean_tensor_name],
        operator.output_full_names,
        container,
        axis=1,
        broadcast=1,
    )


register_converter("meanImagePreprocessor", convert_preprocessing_mean_image)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_scaler(scope, operator, container):
    op_type = "Scaler"
    attrs = {"name": operator.full_name}
    scaler = operator.raw_operator.scaler

    scale = [x for x in scaler.scaleValue]
    offset = [-x for x in scaler.shiftValue]

    attrs["scale"] = scale
    attrs["offset"] = offset

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("scaler", convert_scaler)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_crop_height_width
from ....common._registration import register_converter


def convert_crop(scope, operator, container):
    # Extract number of pixels cropped in CoreML operator.
    border = operator.raw_operator.crop.cropAmounts.borderAmounts

    # Compute cropping amounts. left_border=1 means one pixel will be removed at the beginning
    # of W-axis. bottom_border=1 means one pixel will be removed at the end of H-axis.
    left_border, top_border, right_border, bottom_border = (0,) * 4
    if len(operator.input_full_names) == 1:
        left_border = border[1].startEdgeSize
        top_border = border[0].startEdgeSize
        right_border = border[1].endEdgeSize
        bottom_border = border[0].endEdgeSize
    else:
        offset = operator.raw_operator.crop.offset
        in_shape = operator.inputs[0].type.shape
        out_shape = operator.outputs[0].type.shape
        left_border = offset[1]
        top_border = offset[0]
        right_border = in_shape[3] - left_border - out_shape[3]
        bottom_border = in_shape[2] - top_border - out_shape[2]

    # Delegate the selection of ONNX operator to a version-dependent function.
    apply_crop_height_width(
        scope,
        operator.input_full_names[0],
        operator.output_full_names[0],
        container,
        operator_name=operator.full_name,
        top_border=top_border,
        bottom_border=bottom_border,
        left_border=left_border,
        right_border=right_border,
    )


register_converter("crop", convert_crop)



# SPDX-License-Identifier: Apache-2.0

"""
Helpers to test runtimes.
"""
import warnings

import numpy
from .utils_backend import (
    load_data_and_model,
    extract_options,
    ExpectedAssertionError,
    OnnxRuntimeAssertionError,
    compare_outputs,
)


def compare_runtime(test, decimal=5, options=None, verbose=False, context=None):
    """
    The function compares the expected output (computed with
    the model before being converted to ONNX) and the ONNX output
    produced with module *onnxruntime*.

    :param test: dictionary with the following keys:
        - *onnx*: onnx model (filename or object)
        - *expected*: expected output (filename pkl or object)
        - *data*: input data (filename pkl or object)
    :param decimal: precision of the comparison
    :param options: comparison options
    :param context: specifies custom operators
    :param verbose: in case of error, the function may print
        more information on the standard output

    The function does not return anything but raises an error
    if the comparison failed.
    """
    if context is None:
        context = {}
    load = load_data_and_model(test, **context)

    onx = test["onnx"]
    if options is None:
        if isinstance(onx, str):
            options = extract_options(onx)
        else:
            options = {}
    elif not isinstance(options, dict):
        raise TypeError("options must be a dictionary.")

    try:
        import onnxruntime
    except ImportError:
        warnings.warn("Unable to import onnxruntime.")
        return

    try:
        sess = onnxruntime.InferenceSession(onx, providers=["CPUExecutionProvider"])
    except ExpectedAssertionError as expe:
        raise expe
    except Exception as e:
        if "CannotLoad" in options:
            raise ExpectedAssertionError(
                "Unable to load onnx '{0}' due to\n{1}".format(onx, e)
            )
        else:
            if verbose:
                import onnx

                model = onnx.load(onx)
                smodel = "\nJSON ONNX\n" + str(model)
            else:
                smodel = ""
            raise OnnxRuntimeAssertionError(
                "Unable to load onnx '{0}' due to {1}\nONNX\n{2}".format(onx, e, smodel)
            )

    input = load["data"]
    if isinstance(input, dict):
        inputs = input
    elif isinstance(input, (list, numpy.ndarray)):
        inp = sess.get_inputs()
        if len(inp) == len(input):
            inputs = {i.name: v for i, v in zip(inp, input)}
        elif len(inp) == 1:
            inputs = {inp[0].name: input}
        elif isinstance(input, numpy.ndarray):
            shape = sum(i.shape[1] if len(i.shape) == 2 else i.shape[0] for i in inp)
            if shape == input.shape[1]:
                inputs = {n.name: input[:, i] for i, n in enumerate(inp)}
            else:
                raise OnnxRuntimeAssertionError(
                    "Wrong number of inputs onnx {0} "
                    "!= original shape {1}, onnx='{2}'".format(
                        len(inp), input.shape, onnx
                    )
                )
        elif isinstance(input, list):
            try:
                array_input = numpy.array(input)
            except Exception:
                raise OnnxRuntimeAssertionError(
                    "Wrong number of inputs onnx {0} != original {1}, "
                    "onnx='{2}'".format(len(inp), len(input), onnx)
                )
            shape = sum(i.shape[1] for i in inp)
            if shape == array_input.shape[1]:
                inputs = {
                    n.name: _create_column([row[i] for row in input], n.type)
                    for i, n in enumerate(inp)
                }
            else:
                raise OnnxRuntimeAssertionError(
                    "Wrong number of inputs onnx {0} != original "
                    "shape {1}, onnx='{2}'*".format(len(inp), array_input.shape, onnx)
                )
        else:
            raise OnnxRuntimeAssertionError(
                "Wrong number of inputs onnx {0} != original {1}, onnx='{2}'".format(
                    len(inp), len(input), onnx
                )
            )
    else:
        raise OnnxRuntimeAssertionError(
            "Dict or list is expected, not {0}".format(type(input))
        )

    for k in inputs:
        if isinstance(inputs[k], list):
            inputs[k] = numpy.array(inputs[k])

    OneOff = options.pop("OneOff", False)
    if OneOff:
        if len(inputs) == 1:
            name, values = list(inputs.items())[0]
            res = []
            for input in values:
                try:
                    one = sess.run(None, {name: input})
                except ExpectedAssertionError as expe:
                    raise expe
                except Exception as e:
                    raise OnnxRuntimeAssertionError(
                        "Unable to run onnx '{0}' due to {1}".format(onnx, e)
                    )
                res.append(one)
            output = _post_process_output(res)
        else:

            def to_array(vv):
                if isinstance(vv, (numpy.ndarray, numpy.int64, numpy.float32)):
                    return numpy.array([vv])
                else:
                    return numpy.array([vv], dtype=numpy.float32)

            t = list(inputs.items())[0]
            res = []
            for i in range(0, len(t[1])):
                iii = {k: to_array(v[i]) for k, v in inputs.items()}
                try:
                    one = sess.run(None, iii)
                except ExpectedAssertionError as expe:
                    raise expe
                except Exception as e:
                    raise OnnxRuntimeAssertionError(
                        "Unable to run onnx '{0}' due to {1}".format(onx, e)
                    )
                res.append(one)
            output = _post_process_output(res)
    else:
        try:
            output = sess.run(None, inputs)
        except ExpectedAssertionError as expe:
            raise expe
        except RuntimeError as e:
            if "-Fail" in onx:
                raise ExpectedAssertionError(
                    "onnxruntime cannot compute the prediction for '{0}'".format(onx)
                )
            else:
                raise OnnxRuntimeAssertionError(
                    "onnxruntime cannot compute the prediction for '{0}' due to {1}".format(
                        onx, e
                    )
                )
        except Exception as e:
            raise OnnxRuntimeAssertionError(
                "Unable to run onnx '{0}' due to {1}".format(onx, e)
            )

    output0 = output.copy()

    try:
        _compare_expected(
            load["expected"], output, sess, onx, decimal=decimal, **options
        )
    except ExpectedAssertionError as expe:
        raise expe
    except Exception as e:
        if verbose:
            import onnx

            model = onnx.load(onx)
            smodel = "\nJSON ONNX\n" + str(model)
        else:
            smodel = ""
        raise OnnxRuntimeAssertionError(
            "Model '{0}' has discrepancies.\n{1}: {2}{3}".format(
                onx, type(e), e, smodel
            )
        )

    return output0


def _post_process_output(res):
    """
    Applies post processings before running the comparison
    such as changing type from list to arrays.
    """
    if isinstance(res, list):
        if len(res) == 0:
            return res
        elif len(res) == 1:
            return _post_process_output(res[0])
        elif isinstance(res[0], numpy.ndarray):
            return numpy.array(res)
        elif isinstance(res[0], dict):
            import pandas

            return pandas.DataFrame(res).values
        else:
            ls = [len(r) for r in res]
            mi = min(ls)
            if mi != max(ls):
                raise NotImplementedError(
                    "Unable to postprocess various number of outputs in [{0}, {1}]".format(
                        min(ls), max(ls)
                    )
                )
            if mi > 1:
                output = []
                for i in range(mi):
                    output.append(_post_process_output([r[i] for r in res]))
                return output
            elif isinstance(res[0], list):
                # list of lists
                if isinstance(res[0][0], list):
                    return numpy.array(res)
                elif len(res[0]) == 1 and isinstance(res[0][0], dict):
                    return _post_process_output([r[0] for r in res])
                elif len(res) == 1:
                    return res
                else:
                    if len(res[0]) != 1:
                        raise NotImplementedError(
                            "Not conversion implemented for {0}".format(res)
                        )
                    st = [r[0] for r in res]
                    return numpy.vstack(st)
            else:
                return res
    else:
        return res


def _create_column(values, dtype):
    "Creates a column from values with dtype"
    if str(dtype) == "tensor(int64)":
        return numpy.array(values, dtype=numpy.int64)
    elif str(dtype) == "tensor(float)":
        return numpy.array(values, dtype=numpy.float32)
    else:
        raise OnnxRuntimeAssertionError(
            "Unable to create one column from dtype '{0}'".format(dtype)
        )


def _compare_expected(
    expected, output, sess, onnx, decimal=5, onnx_shape=None, **kwargs
):
    """
    Compares the expected output against the runtime outputs.
    This is specific to *onnxruntime* due to variable *sess*
    of type *onnxruntime.InferenceSession*.
    """
    tested = 0
    if isinstance(expected, list):
        if isinstance(output, list):
            if "Out0" in kwargs:
                expected = expected[:1]
                output = output[:1]
                del kwargs["Out0"]
            if "Reshape" in kwargs:
                del kwargs["Reshape"]
                output = numpy.hstack(output).ravel()
                output = output.reshape(
                    (len(expected), len(output.ravel()) // len(expected))
                )
            if len(expected) != len(output):
                raise OnnxRuntimeAssertionError(
                    "Unexpected number of outputs '{0}', expected={1}, got={2}".format(
                        onnx, len(expected), len(output)
                    )
                )
            try:
                onnx_shapes = [_.shape for _ in sess.get_outputs()]
            except TypeError:
                # Unable to convert function return value to a Python type!
                onnx_shapes = [None for o in output]
            for exp, out, osh in zip(expected, output, onnx_shapes):
                _compare_expected(
                    exp, out, sess, onnx, decimal=decimal, onnx_shape=osh, **kwargs
                )
                tested += 1
        else:
            raise OnnxRuntimeAssertionError(
                "Type mismatch for '{0}', output type is {1}".format(onnx, type(output))
            )
    elif isinstance(expected, dict):
        if not isinstance(output, dict):
            raise OnnxRuntimeAssertionError("Type mismatch for '{0}'".format(onnx))
        for k, v in output.items():
            if k not in expected:
                continue
            msg = compare_outputs(expected[k], v, decimal=decimal, **kwargs)
            if msg:
                raise OnnxRuntimeAssertionError(
                    "Unexpected output '{0}' in model '{1}'\n{2}".format(k, onnx, msg)
                )
            tested += 1
    elif isinstance(expected, numpy.ndarray):
        if isinstance(output, list):
            if expected.shape[0] == len(output) and isinstance(output[0], dict):
                import pandas

                output = pandas.DataFrame(output)
                output = output[list(sorted(output.columns))]
                output = output.values
        if isinstance(output, (dict, list)):
            if len(output) != 1:
                ex = str(output)
                if len(ex) > 70:
                    ex = ex[:70] + "..."
                raise OnnxRuntimeAssertionError(
                    "More than one output when 1 is expected for onnx '{0}'"
                    "\n{1}".format(onnx, ex)
                )
            output = output[-1]
        if not isinstance(output, numpy.ndarray):
            raise OnnxRuntimeAssertionError(
                "output must be an array for onnx '{0}' not {1}".format(
                    onnx, type(output)
                )
            )
        if onnx_shape is not None:
            if len(onnx_shape) == 2 and onnx_shape[0] is not None:
                cols = onnx_shape[1]
                ecols = output.shape[1] if len(output.shape) == 2 else 1
                if cols != ecols:
                    raise OnnxRuntimeAssertionError(
                        "Unexpected onnx shape {0} != {1} for onnx '{2}'".format(
                            onnx_shape, output.shape, onnx
                        )
                    )
        msg = compare_outputs(expected, output, decimal=decimal, **kwargs)
        if isinstance(msg, ExpectedAssertionError):
            raise msg
        if msg:
            raise OnnxRuntimeAssertionError(
                "Unexpected output in model '{0}'\n{1}".format(onnx, msg)
            )
        tested += 1
    else:
        from scipy.sparse.csr import csr_matrix

        if isinstance(expected, csr_matrix):
            # DictVectorizer
            one_array = numpy.array(output)
            msg = compare_outputs(
                expected.todense(), one_array, decimal=decimal, **kwargs
            )
            if msg:
                raise OnnxRuntimeAssertionError(
                    "Unexpected output in model '{0}'\n{1}".format(onnx, msg)
                )
            tested += 1
        else:
            raise OnnxRuntimeAssertionError(
                "Unexpected type for expected output ({1}) and onnx '{0}'".format(
                    onnx, type(expected)
                )
            )
    if tested == 0:
        raise OnnxRuntimeAssertionError("No test for onnx '{0}'".format(onnx))


def run_with_runtime(inputs, model_path):
    """

    :param inputs: inputs to the model
    :param model_path: onnx model file path
    :return: (output,session)
    """
    try:
        import onnxruntime

        session = onnxruntime.InferenceSession(
            model_path, providers=["CPUExecutionProvider"]
        )
        output = session.run(None, inputs)
        return (output, session)
    except Exception as e:
        raise OnnxRuntimeAssertionError(
            "The runtime does either not exists of fails to load model"
        ) from e



# SPDX-License-Identifier: Apache-2.0

import unittest
import numpy as np
from onnx import TensorProto
from onnx.defs import onnx_opset_version
from onnx.helper import (
    make_node,
    make_graph,
    make_model,
    make_tensor_value_info,
    make_opsetid,
)
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxruntime import InferenceSession
from onnxmltools.convert.sparkml.operator_converters.tree_helper import Node

TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmDecisionTreeClassifierBig(unittest.TestCase):
    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split(self):
        attrs = {
            "class_ids": [
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
            ],
            "class_nodeids": [
                3,
                3,
                3,
                4,
                4,
                4,
                6,
                6,
                6,
                7,
                7,
                7,
                10,
                10,
                10,
                11,
                11,
                11,
                12,
                12,
                12,
            ],
            "class_treeids": [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ],
            "class_weights": [
                0.4,
                0.6,
                0.0,
                0.95,
                0.04,
                0.0,
                0.0,
                0.185,
                0.814,
                0.372,
                0.62,
                0.0,
                0.74,
                0.21,
                0.03,
                0.0,
                1.0,
                0.0,
                0.87,
                0.05,
                0.071,
            ],
            "classlabels_int64s": [0, 1, 2],
            "name": "TreeEnsembleClassifier",
            "nodes_falsenodeids": [8, 5, 4, 0, 0, 7, 0, 0, 12, 11, 0, 0, 0],
            "nodes_featureids": [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            "nodes_hitrates": [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
            ],
            "nodes_missing_value_tracks_true": [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            "nodes_modes": [
                "BRANCH_LEQ",
                "||",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "LEAF",
            ],
            "nodes_nodeids": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            "nodes_treeids": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            "nodes_truenodeids": [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            "nodes_values": [
                10.5,
                [55, 59, 65],
                1.5,
                0.0,
                0.0,
                8.5,
                0.0,
                0.0,
                10.5,
                9.5,
                0.0,
                0.0,
                0.0,
            ],
            "post_transform": None,
        }
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n"
                    f"{attrs['nodes_modes']}\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        node = make_node(
            op_type="TreeEnsembleClassifier",
            inputs=["X"],
            outputs=["L", "Y"],
            **new_attrs,
        )
        X = make_tensor_value_info("X", TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info("Y", TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info("L", TensorProto.INT64, [None])
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(
            model.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {"X": x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [2, 2, 2, 2, 2, 1, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        self.assertEqual(new_attrs, new_new_attrs)

    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split_non_contiguous_ids(self):
        attrs = {
            "class_ids": [
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
            ],
            "class_nodeids": [
                3,
                3,
                3,
                4,
                4,
                4,
                6,
                6,
                6,
                7,
                7,
                7,
                10,
                10,
                10,
                11,
                11,
                11,
                13,
                13,
                13,
            ],
            "class_treeids": [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ],
            "class_weights": [
                0.4,
                0.6,
                0.0,
                0.95,
                0.04,
                0.0,
                0.0,
                0.185,
                0.814,
                0.372,
                0.62,
                0.0,
                0.74,
                0.21,
                0.03,
                0.0,
                1.0,
                0.0,
                0.87,
                0.05,
                0.071,
            ],
            "classlabels_int64s": [0, 1, 2],
            "name": "TreeEnsembleClassifier",
            "nodes_falsenodeids": [8, 5, 4, 0, 0, 7, 0, 0, 13, 11, 0, 0, 0],
            "nodes_featureids": [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            "nodes_hitrates": [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
            ],
            "nodes_missing_value_tracks_true": [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            "nodes_modes": [
                "BRANCH_LEQ",
                "||",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "LEAF",
            ],
            "nodes_nodeids": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13],
            "nodes_treeids": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            "nodes_truenodeids": [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            "nodes_values": [
                10.5,
                [55, 59, 65],
                1.5,
                0.0,
                0.0,
                8.5,
                0.0,
                0.0,
                10.5,
                9.5,
                0.0,
                0.0,
                0.0,
            ],
            "post_transform": None,
        }
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n"
                    f"{attrs['nodes_modes']}\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        node = make_node(
            op_type="TreeEnsembleClassifier",
            inputs=["X"],
            outputs=["L", "Y"],
            **new_attrs,
        )
        X = make_tensor_value_info("X", TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info("Y", TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info("L", TensorProto.INT64, [None])
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(
            model.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {"X": x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [2, 2, 2, 2, 2, 1, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        self.assertEqual(new_attrs, new_new_attrs)

    @unittest.skipIf(TARGET_OPSET < 17, reason="Opset 17 is needed")
    def test_split_more_complex(self):
        attrs = {
            "class_ids": [
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
                0,
                1,
                2,
            ],
            "class_nodeids": [
                3,
                3,
                3,
                4,
                4,
                4,
                6,
                6,
                6,
                7,
                7,
                7,
                10,
                10,
                10,
                11,
                11,
                11,
                12,
                12,
                12,
            ],
            "class_treeids": [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ],
            "class_weights": [
                0.4,
                0.6,
                0.0,
                0.95,
                0.04,
                0.0,
                0.0,
                0.185,
                0.814,
                0.372,
                0.62,
                0.0,
                0.74,
                0.21,
                0.03,
                0.0,
                1.0,
                0.0,
                0.87,
                0.05,
                0.071,
            ],
            "classlabels_int64s": [0, 1, 2],
            "name": "TreeEnsembleClassifier",
            "nodes_falsenodeids": [8, 5, 4, 0, 0, 7, 0, 0, 12, 11, 0, 0, 0],
            "nodes_featureids": [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
            "nodes_hitrates": [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
            ],
            "nodes_missing_value_tracks_true": [
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
            "nodes_modes": [
                "BRANCH_LEQ",
                "BRANCH_LEQ",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "BRANCH_LEQ",
                "BRANCH_LEQ",
                "LEAF",
                "LEAF",
                "LEAF",
            ],
            "nodes_nodeids": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            "nodes_treeids": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            "nodes_truenodeids": [1, 2, 3, 0, 0, 6, 0, 0, 9, 10, 0, 0, 0],
            "nodes_values": [
                [100, 101, 102],
                [55, 59, 65],
                [10, 11, 12],
                0.0,
                0.0,
                [1000, 1001, 1002],
                0.0,
                0.0,
                [10000, 10001],
                [100000, 100001],
                0.0,
                0.0,
                0.0,
            ],
            "post_transform": None,
        }
        root, nodes = Node.create(attrs)
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
        n_nodes1 = len(list(root))
        root.unfold_rule_or()
        n_nodes2 = len(list(root))
        if n_nodes1 >= n_nodes2:
            raise AssertionError(f"Unexpected {n_nodes1} >= {n_nodes2}")
        ns = {}
        for node in root:
            key = node.nodes_treeids, node.nodes_nodeids
            if None in key:
                raise AssertionError(f"Wrong key for node {node!r}.")
            if key in ns:
                raise AssertionError(f"Duplicate node id {key}.")
            ns[key] = node

        # back to onnx
        new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{attrs['nodes_modes']}"
                    f"\nbefore {attrs[k]!r}"
                    f"\nafter  {new_attrs[k]!r}\n{new_attrs['nodes_modes']}"
                )

        self.assertEqual(len(new_attrs["nodes_modes"]), len(attrs["nodes_modes"]) + 10)
        node = make_node(
            op_type="TreeEnsembleClassifier",
            inputs=["X"],
            outputs=["L", "Y"],
            **new_attrs,
        )
        X = make_tensor_value_info("X", TensorProto.FLOAT, [None, None])
        Y = make_tensor_value_info("Y", TensorProto.FLOAT, [None, None])
        L = make_tensor_value_info("L", TensorProto.INT64, [None])
        graph = make_graph([node], "n", [X], [L, Y])
        opset_imports = [make_opsetid("", 17), make_opsetid("ai.onnx.ml", 3)]
        model = make_model(graph, opset_imports=opset_imports)
        sess = InferenceSession(
            model.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        x = np.arange(20).reshape((-1, 2)).astype(np.float32)
        got = sess.run(None, {"X": x})
        self.assertEqual(len(got), 2)
        self.assertEqual(got[0].tolist(), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

        # again
        root, nodes = Node.create(new_attrs)
        root.unfold_rule_or()
        new_new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        self.assertEqual(new_attrs, new_new_attrs)

    def test_debug(self):
        try:
            from onnxmltools.debug1 import attrs
        except ImportError:
            return
        root, nodes = Node.create(attrs)
        root.unfold_rule_or()
        new_attrs = root.to_attrs(
            post_transform=None,
            name="TreeEnsembleClassifier",
            classlabels_int64s=[0, 1, 2],
            domain="ai.onnx.ml",
        )
        for k in attrs:
            if k in {"post_transform"}:
                continue
            if len(attrs[k]) > len(new_attrs[k]):
                raise AssertionError(
                    f"Issue with attribute {k!r}\n{len(new_attrs[k])}"
                    f"\nbefore {len(attrs[k])}."
                )


if __name__ == "__main__":
    # TestSparkmDecisionTreeClassifierBig().test_debug()
    unittest.main(verbosity=2)



from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._topology import Operator, Scope, ModelComponentContainer
from ....proto import onnx_proto
from pyspark.ml.clustering import KMeansModel
import numpy as np


def convert_sparkml_k_means_model(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    if container.target_opset < 7:
        raise NotImplementedError(
            "Converting to ONNX for KMeansModel is not supported in opset < 7"
        )

    op: KMeansModel = operator.raw_operator
    centers: np.ndarray = np.vstack(op.clusterCenters())

    K = centers.shape[0]  # number of clusters
    C = operator.inputs[0].type.shape[1]  # Number of features from input

    if centers.shape[1] != C:
        raise ValueError(
            f"Number of features {centers.shape[1]} "
            f"in input does not match number of features in centers {C}"
        )

    # [K x C]
    centers_variable_name = scope.get_unique_variable_name("centers")
    container.add_initializer(
        centers_variable_name,
        onnx_proto.TensorProto.FLOAT,
        centers.shape,
        centers.flatten().astype(np.float32),
    )

    distance_output_variable_name = scope.get_unique_variable_name("distance_output")

    if op.getDistanceMeasure() == "euclidean":
        # [1 x K]
        centers_row_squared_sum_variable_name = scope.get_unique_variable_name(
            "centers_row_squared_sum"
        )
        centers_row_squared_sum = (
            np.sum(centers**2, axis=-1).flatten().astype(np.float32)
        )
        container.add_initializer(
            centers_row_squared_sum_variable_name,
            onnx_proto.TensorProto.FLOAT,
            [1, K],
            centers_row_squared_sum,
        )

        # input_row_squared_sum: [N x 1]
        input_row_squared_sum_variable_name = scope.get_unique_variable_name(
            "input_row_squared_sum"
        )
        reduce_sum_square_attrs = {
            "name": scope.get_unique_operator_name("input_row_squared_sum"),
            "axes": [1],
            "keepdims": 1,
        }
        container.add_node(
            op_type="ReduceSumSquare",
            inputs=[operator.inputs[0].full_name],
            outputs=[input_row_squared_sum_variable_name],
            **reduce_sum_square_attrs,
        )

        # -2 * input * Transpose(Center) + input_row_squared_sum: [N x K]
        gemm_output_variable_name = scope.get_unique_variable_name("gemm_output")
        gemm_attrs = {
            "name": scope.get_unique_operator_name("GeMM"),
            "alpha": -2.0,
            "beta": 1.0,
            "transB": 1,
        }
        container.add_node(
            op_type="Gemm",
            inputs=[
                operator.inputs[0].full_name,
                centers_variable_name,
                input_row_squared_sum_variable_name,
            ],
            outputs=[gemm_output_variable_name],
            op_version=7,
            **gemm_attrs,
        )

        # Euclidean Distance Squared = input_row_squared_sum - 2 *
        # input * Transpose(Center) + Transpose(centers_row_squared_sum)
        # [N x K]
        container.add_node(
            op_type="Add",
            inputs=[gemm_output_variable_name, centers_row_squared_sum_variable_name],
            outputs=[distance_output_variable_name],
            op_version=7,
        )
    elif op.getDistanceMeasure() == "cosine":
        # centers_row_norm2: [1 x K]
        centers_row_norm2_variable_name = scope.get_unique_variable_name(
            "centers_row_norm2"
        )
        centers_row_norm2 = (
            np.linalg.norm(centers, ord=2, axis=1).flatten().astype(np.float32)
        )
        container.add_initializer(
            centers_row_norm2_variable_name,
            onnx_proto.TensorProto.FLOAT,
            [1, K],
            centers_row_norm2,
        )

        # input_row_norm2: [N x 1]
        input_row_norm2_variable_name = scope.get_unique_variable_name(
            "input_row_norm2"
        )
        reduce_l2_attrs = {
            "name": scope.get_unique_operator_name("input_row_norm2"),
            "axes": [1],
            "keepdims": 1,
        }
        container.add_node(
            op_type="ReduceL2",
            inputs=[operator.inputs[0].full_name],
            outputs=[input_row_norm2_variable_name],
            **reduce_l2_attrs,
        )

        # input * Transpose(Center): [N x K]
        zeros_variable_name = scope.get_unique_variable_name("zeros")
        container.add_initializer(
            zeros_variable_name,
            onnx_proto.TensorProto.FLOAT,
            [1, K],
            np.zeros([1, K]).flatten().astype(np.float32),
        )
        gemm_output_variable_name = scope.get_unique_variable_name("gemm_output")
        gemm_attrs = {
            "name": scope.get_unique_operator_name("GeMM"),
            "alpha": 1.0,
            "transB": 1,
        }
        container.add_node(
            op_type="Gemm",
            inputs=[
                operator.inputs[0].full_name,
                centers_variable_name,
                zeros_variable_name,
            ],
            outputs=[gemm_output_variable_name],
            op_version=7,
            **gemm_attrs,
        )

        # Cosine similarity = gemm_output / input_row_norm2 / centers_row_norm2: [N x K]
        div_output_variable_name = scope.get_unique_variable_name("div_output")
        container.add_node(
            op_type="Div",
            inputs=[gemm_output_variable_name, input_row_norm2_variable_name],
            outputs=[div_output_variable_name],
            op_version=7,
        )
        cosine_similarity_output_variable_name = scope.get_unique_variable_name(
            "cosine_similarity_output"
        )
        container.add_node(
            op_type="Div",
            inputs=[div_output_variable_name, centers_row_norm2_variable_name],
            outputs=[cosine_similarity_output_variable_name],
            op_version=7,
        )

        # Cosine distance - 1 = -Cosine similarity: [N x K]
        container.add_node(
            op_type="Neg",
            inputs=[cosine_similarity_output_variable_name],
            outputs=[distance_output_variable_name],
        )
    else:
        raise ValueError(f"Distance measure {op.getDistanceMeasure()} not supported")

    # ArgMin(distance): [N]
    argmin_attrs = {
        "axis": 1,
        "keepdims": 0,
    }
    container.add_node(
        op_type="ArgMin",
        inputs=[distance_output_variable_name],
        outputs=[operator.outputs[0].full_name],
        **argmin_attrs,
    )


register_converter("pyspark.ml.clustering.KMeansModel", convert_sparkml_k_means_model)


def calculate_k_means_model_output_shapes(operator: Operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Input must be a [N, C]-tensor")

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = Int64TensorType(shape=[N])


register_shape_calculator(
    "pyspark.ml.clustering.KMeansModel", calculate_k_means_model_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_mean
from ....common._registration import register_converter


def convert_average(scope, operator, container):
    apply_mean(
        scope,
        operator.input_full_names,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
    )


register_converter("average", convert_average)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import packaging.version as pv
import inspect
import os
import onnx
import pandas
import numpy
from pyspark.ml.classification import LogisticRegression, OneVsRest
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmOneVsRest(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    @unittest.skipIf(
        pv.Version(onnx.__version__) <= pv.Version("1.3"), "Need Greater Opset 9"
    )
    def test_one_vs_rest(self):
        this_script_dir = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        input_path = os.path.join(
            this_script_dir, "data", "sample_multiclass_classification_data.txt"
        )
        data = self.spark.read.format("libsvm").load(input_path)
        lr = LogisticRegression(maxIter=100, tol=0.0001, regParam=0.01)
        ovr = OneVsRest(classifier=lr)
        model = ovr.fit(data)

        feature_count = data.first()[1].size
        model_onnx = convert_sparkml(
            model,
            "Sparkml OneVsRest",
            [("features", FloatTensorType([None, feature_count]))],
            spark_session=self.spark,
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        data_np = (
            data.toPandas()
            .features.apply(lambda x: pandas.Series(x.toArray()))
            .values.astype(numpy.float32)
        )
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
        ]
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlOneVsRest"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["prediction"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_dot_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C], [N, C] ---> [N, 1]
        2. [N, C, 1, 1], [N, C, 1, 1] ---> [N, 1, 1, 1]
    """
    check_input_and_output_numbers(operator, input_count_range=2, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    if operator.inputs[0].type.shape != operator.inputs[1].type.shape:
        raise RuntimeError("Input shapes must be identical")

    # Assume that inputs are [N, C]- or [N, C, 1, 1]-tensors
    output_shape = copy.deepcopy(operator.inputs[0].type.shape)
    output_shape[1] = 1
    operator.outputs[0].type.shape = output_shape


register_shape_calculator("dot", calculate_dot_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_tile
from ....common._registration import register_converter


def convert_sequence_repeat(scope, operator, container):
    repeat_count = (
        operator.raw_operator.sequenceRepeat.nRepetitions
    )  # number of copies along N-axis in CoreML

    if len(operator.inputs[0].type.shape) == 4:
        # Number of copies along [N, C, H, W]
        repeats = [int(repeat_count), 1, 1, 1]
    elif len(operator.inputs[0].type.shape) == 2:
        # Number of copies along [N, C]
        repeats = [int(repeat_count), 1]

    apply_tile(
        scope,
        operator.input_full_names[0],
        operator.output_full_names[0],
        container,
        operator_name=operator.full_name,
        repeats=repeats,
    )


register_converter("sequenceRepeat", convert_sequence_repeat)



# SPDX-License-Identifier: Apache-2.0

import collections

from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers


def convert_sparkml_linear_regressor(scope, operator, container):
    op = operator.raw_operator
    op_type = "LinearRegressor"
    attrs = {
        "name": scope.get_unique_operator_name(op_type),
        "coefficients": op.coefficients.astype(float),
        "intercepts": (
            op.intercept.astype(float)
            if isinstance(op.intercept, collections.abc.Iterable)
            else [float(op.intercept)]
        ),
    }
    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter(
    "pyspark.ml.regression.LinearRegressionModel", convert_sparkml_linear_regressor
)
register_converter(
    "pyspark.ml.regression.GeneralizedLinearRegressionModel",
    convert_sparkml_linear_regressor,
)


def calculate_linear_regressor_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, 1])


register_shape_calculator(
    "pyspark.ml.regression.LinearRegressionModel",
    calculate_linear_regressor_output_shapes,
)
register_shape_calculator(
    "pyspark.ml.regression.GeneralizedLinearRegressionModel",
    calculate_linear_regressor_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import numpy

from ..common._container import LightGbmModelContainer
from ..common._topology import Topology
from ..common.data_types import (
    FloatTensorType,
    SequenceType,
    DictionaryType,
    StringType,
    Int64Type,
)

from lightgbm import LGBMClassifier, LGBMRegressor

lightgbm_classifier_list = [LGBMClassifier]

# Associate scikit-learn types with our operator names.
# If two scikit-learn models share a single name, it means their
# are equivalent in terms of conversion.
lightgbm_operator_name_map = {
    LGBMClassifier: "LgbmClassifier",
    LGBMRegressor: "LgbmRegressor",
}


class WrappedBooster:
    def __init__(self, booster):
        self.booster_ = booster
        self.n_features_ = self.booster_.feature_name()
        self.objective_ = self.get_objective()
        if self.objective_.startswith("binary"):
            self.operator_name = "LgbmClassifier"
            self.classes_ = self._generate_classes(booster)
        elif self.objective_.startswith("multiclass"):
            self.operator_name = "LgbmClassifier"
            self.classes_ = self._generate_classes(booster)
        elif self.objective_.startswith("regression"):
            self.operator_name = "LgbmRegressor"
        else:
            raise NotImplementedError(
                "Unsupported LightGbm objective: %r." % self.objective_
            )
        try:
            average_output = self.booster_.attr("average_output")
        except AttributeError:
            average_output = self.booster_.params.get("average_output", None)
        if average_output:
            self.boosting_type = "rf"
        else:
            # Other than random forest, other boosting types do not affect later conversion.
            # Here `gbdt` is chosen for no reason.
            self.boosting_type = "gbdt"

    @staticmethod
    def _generate_classes(booster):
        if isinstance(booster, dict):
            num_class = booster["num_class"]
        else:
            try:
                num_class = booster.attr("num_class")
            except AttributeError:
                num_class = booster.params.get("num_class", None)
        if num_class is None:
            dp = booster.dump_model(num_iteration=1)
            num_class = dp["num_class"]
        if num_class == 1:
            return numpy.asarray([0, 1])
        return numpy.arange(num_class)

    def get_objective(self):
        "Returns the objective."
        if hasattr(self, "objective_") and self.objective_ is not None:
            return self.objective_
        try:
            objective = self.booster_.attr("objective")
        except AttributeError:
            objective = self.booster_.params.get("objective", None)
        if objective is not None:
            return objective
        dp = self.booster_.dump_model(num_iteration=1)
        return dp["objective"]


def _get_lightgbm_operator_name(model):
    """
    Get operator name of the input argument

    :param model:  A lightgbm object.
    :return: A string which stands for the type of the input model in our conversion framework
    """
    if isinstance(model, WrappedBooster):
        return model.operator_name
    model_type = type(model)
    if model_type not in lightgbm_operator_name_map:
        raise ValueError("No proper operator name found for '%s'" % model_type)
    return lightgbm_operator_name_map[model_type]


def _parse_lightgbm_simple_model(scope, model, inputs, split=None):
    """
    This function handles all non-pipeline models.

    :param scope: Scope object
    :param model: A lightgbm object
    :param inputs: A list of variables
    :param split: split TreeEnsembleRegressor into multiple node to reduce
        discrepancies
    :return: A list of output variables which will be passed to next stage
    """
    operator_name = _get_lightgbm_operator_name(model)
    this_operator = scope.declare_local_operator(operator_name, model)
    this_operator.split = split
    this_operator.inputs = inputs

    if operator_name == "LgbmClassifier":
        # For classifiers, we may have two outputs, one for
        # label and the other one for probabilities of all classes.
        # Notice that their types here are not necessarily correct
        # and they will be fixed in shape inference phase
        label_variable = scope.declare_local_variable("lgbmlabel", FloatTensorType())
        probability_map_variable = scope.declare_local_variable(
            "lgbmprobabilities", FloatTensorType()
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(probability_map_variable)
    else:
        # We assume that all scikit-learn operator can only produce a single float tensor.
        variable = scope.declare_local_variable("variable", FloatTensorType())
        this_operator.outputs.append(variable)
    return this_operator.outputs


def _parse_sklearn_classifier(scope, model, inputs, zipmap=True):
    probability_tensor = _parse_lightgbm_simple_model(scope, model, inputs)
    this_operator = scope.declare_local_operator("LgbmZipMap")
    this_operator.inputs = probability_tensor
    this_operator.zipmap = zipmap

    classes = model.classes_
    label_type = Int64Type()

    if isinstance(model.classes_, list) and isinstance(
        model.classes_[0], numpy.ndarray
    ):
        # multi-label problem
        # this_operator.classlabels_int64s = list(range(0, len(classes)))
        raise NotImplementedError("multi-label is not supported")
    elif numpy.issubdtype(model.classes_.dtype, numpy.floating):
        classes = numpy.array(list(map(lambda x: int(x), classes)))
        if set(map(lambda x: float(x), classes)) != set(model.classes_):
            raise RuntimeError(
                "skl2onnx implicitly converts float class "
                "labels into integers but at least one label "
                "is not an integer. Class labels should "
                "be integers or strings."
            )
        this_operator.classlabels_int64s = classes
    elif numpy.issubdtype(model.classes_.dtype, numpy.signedinteger):
        this_operator.classlabels_int64s = classes
    else:
        classes = numpy.array([s.encode("utf-8") for s in classes])
        this_operator.classlabels_strings = classes
        label_type = StringType()

    output_label = scope.declare_local_variable("label", label_type)
    if zipmap:
        output_probability = scope.declare_local_variable(
            "probabilities", SequenceType(DictionaryType(label_type, FloatTensorType()))
        )
    else:
        output_probability = scope.declare_local_variable(
            "probabilities", FloatTensorType(shape=[None, len(classes)])
        )
    this_operator.outputs.append(output_label)
    this_operator.outputs.append(output_probability)
    return this_operator.outputs


def _parse_lightgbm(scope, model, inputs, zipmap=True, split=None):
    """
    This is a delegate function. It doesn't nothing but
    invoke the correct parsing function according to the input
    model's type.

    :param scope: Scope object
    :param model: A lightgbm object
    :param inputs: A list of variables
    :param zipmap: add operator ZipMap after operator TreeEnsembleClassifier
    :param split: split TreeEnsembleRegressor into multiple node to reduce
        discrepancies
    :return: The output variables produced by the input model
    """
    if isinstance(model, LGBMClassifier):
        return _parse_sklearn_classifier(scope, model, inputs, zipmap=zipmap)
    if isinstance(model, WrappedBooster) and model.operator_name == "LgbmClassifier":
        return _parse_sklearn_classifier(scope, model, inputs, zipmap=zipmap)
    return _parse_lightgbm_simple_model(scope, model, inputs, split=split)


def parse_lightgbm(
    model,
    initial_types=None,
    target_opset=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    zipmap=True,
    split=None,
):
    raw_model_container = LightGbmModelContainer(model)
    topology = Topology(
        raw_model_container,
        default_batch_size="None",
        initial_types=initial_types,
        target_opset=target_opset,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
    )
    scope = topology.declare_scope("__root__")

    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    for variable in inputs:
        raw_model_container.add_input(variable)

    outputs = _parse_lightgbm(scope, model, inputs, zipmap=zipmap, split=split)

    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._apply_operation import apply_matmul
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_pca(scope, operator, container):
    op = operator.raw_operator
    pc_tensor = scope.get_unique_variable_name("pc_tensor")
    container.add_initializer(
        pc_tensor,
        onnx_proto.TensorProto.FLOAT,
        [op.pc.numRows, op.pc.numCols],
        list(op.pc.toArray().flatten()),
    )
    apply_matmul(
        scope,
        [operator.inputs[0].full_name, pc_tensor],
        operator.output_full_names,
        container,
    )


register_converter("pyspark.ml.feature.PCAModel", convert_sparkml_pca)


def calculate_sparkml_pca_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType(
        [N, operator.raw_operator.getOrDefault("k")]
    )


register_shape_calculator(
    "pyspark.ml.feature.PCAModel", calculate_sparkml_pca_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter

COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE = {
    0: "BRANCH_LEQ",
    1: "BRANCH_LT",
    2: "BRANCH_GTE",
    3: "BRANCH_GT",
    4: "BRANCH_EQ",
    5: "BRANCH_NEQ",
    6: "LEAF",
}

COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM = {
    0: "NONE",
    1: "SOFTMAX",
    2: "LOGISTIC",
    3: "SOFTMAX_ZERO",
}


def get_onnx_tree_mode(cm_tree_behavior):
    if cm_tree_behavior in COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE:
        return COREML_TREE_NODE_BEHAVIOR_TO_ONNX_TREE_NODE_MODE[cm_tree_behavior]
    raise ValueError(
        "CoreML tree node behavior not supported {0}".format(cm_tree_behavior)
    )


def get_onnx_tree_post_transform(cm_tree_post_transform):
    if cm_tree_post_transform in COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM:
        return COREML_TREE_POST_TRANSFORM_TO_ONNX_TREE_POST_TRANSFORM[
            cm_tree_post_transform
        ]
    raise ValueError(
        "CoreML tree post transform not supported {0}".format(cm_tree_post_transform)
    )


def convert_tree_ensemble_model(scope, operator, container):
    raw_model = operator.raw_operator
    attrs = {"name": operator.full_name}
    if raw_model.WhichOneof("Type") == "treeEnsembleClassifier":
        op_type = "TreeEnsembleClassifier"
        prefix = "class"
        nodes = raw_model.treeEnsembleClassifier.treeEnsemble.nodes
        attrs[
            "base_values"
        ] = raw_model.treeEnsembleClassifier.treeEnsemble.basePredictionValue
        attrs["post_transform"] = get_onnx_tree_post_transform(
            raw_model.treeEnsembleClassifier.postEvaluationTransform
        )
        zipmap_attrs = {"name": scope.get_unique_operator_name("ZipMap")}
        if (
            raw_model.treeEnsembleClassifier.WhichOneof("ClassLabels")
            == "int64ClassLabels"
        ):
            class_labels = list(
                int(i) for i in raw_model.treeEnsembleClassifier.int64ClassLabels.vector
            )
            attrs["classlabels_int64s"] = class_labels
            zipmap_attrs["classlabels_int64s"] = class_labels
        else:
            class_labels = list(
                s.encode("utf-8")
                for s in raw_model.treeEnsembleClassifier.stringClassLabels.vector
            )
            attrs["classlabels_strings"] = class_labels
            zipmap_attrs["classlabels_strings"] = class_labels
    elif raw_model.WhichOneof("Type") == "treeEnsembleRegressor":
        op_type = "TreeEnsembleRegressor"
        prefix = "target"
        nodes = raw_model.treeEnsembleRegressor.treeEnsemble.nodes
        attrs[
            "base_values"
        ] = raw_model.treeEnsembleRegressor.treeEnsemble.basePredictionValue
        attrs[
            "n_targets"
        ] = raw_model.treeEnsembleRegressor.treeEnsemble.numPredictionDimensions
        attrs["post_transform"] = get_onnx_tree_post_transform(
            raw_model.treeEnsembleRegressor.postEvaluationTransform
        )
    else:
        raise ValueError("Unknown tree model type")

    leaf_treeids = [
        node.treeId
        for node in nodes
        if 6 == node.nodeBehavior
        for _ in node.evaluationInfo
    ]
    leaf_nodeids = [
        node.nodeId
        for node in nodes
        if 6 == node.nodeBehavior
        for _ in node.evaluationInfo
    ]
    leaf_ids = [
        weight.evaluationIndex
        for node in nodes
        if 6 == node.nodeBehavior
        for weight in node.evaluationInfo
    ]

    leaf_weights = [
        weight.evaluationValue
        for node in nodes
        if 6 == node.nodeBehavior
        for weight in node.evaluationInfo
    ]

    assert len(leaf_ids) == len(leaf_weights)
    assert len(leaf_weights) == len(leaf_nodeids)
    assert len(leaf_nodeids) == len(leaf_treeids)

    nodes_nodeids = [x.nodeId for x in nodes]
    nodes_treeids = [x.treeId for x in nodes]
    nodes_featureids = [x.branchFeatureIndex for x in nodes]
    nodes_values = [x.branchFeatureValue for x in nodes]
    nodes_truenodeids = [x.trueChildNodeId for x in nodes]
    nodes_falsenodeids = [x.falseChildNodeId for x in nodes]
    nodes_missing_value_tracks_true = [x.missingValueTracksTrueChild for x in nodes]
    nodes_hitrates = [float(x.relativeHitRate) for x in nodes]
    nodes_modes = [get_onnx_tree_mode(x.nodeBehavior) for x in nodes]

    attrs["nodes_treeids"] = nodes_treeids
    attrs["nodes_nodeids"] = nodes_nodeids
    attrs["nodes_featureids"] = nodes_featureids
    attrs["nodes_values"] = nodes_values
    attrs["nodes_hitrates"] = nodes_hitrates
    attrs["nodes_modes"] = nodes_modes
    attrs["nodes_truenodeids"] = nodes_truenodeids
    attrs["nodes_falsenodeids"] = nodes_falsenodeids
    attrs["nodes_missing_value_tracks_true"] = nodes_missing_value_tracks_true
    attrs[prefix + "_treeids"] = leaf_treeids
    attrs[prefix + "_nodeids"] = leaf_nodeids
    attrs[prefix + "_ids"] = leaf_ids
    attrs[prefix + "_weights"] = leaf_weights

    # For regression, we can simply construct a model. For classifier, due to the different representation of
    # classes' probabilities, we need to add some operators for type conversion.
    if raw_model.WhichOneof("Type") == "treeEnsembleRegressor":
        # Create ONNX representation of this operator. If there is only one input, its full topology is
        #
        # input features ---> TreeEnsembleRegressor ---> output
        #
        # If there are multiple (e.g., "N" features) input features, we need to concatenate them all together before feeding them into
        # ONNX tree-based model. It leads to the following computational graph.
        #
        # input feature 1 -----.
        #        ...           |
        #        ...           v
        #        ...      ---> Feature Vectorizer ---> TreeEnsembleRegressor ---> output
        #        ...           ^
        #        ...           |
        # input feature N -----'
        if len(operator.inputs) > 1:
            feature_vector_name = scope.get_unique_variable_name("feature_vector")
            container.add_node(
                "FeatureVectorizer",
                operator.input_full_names,
                feature_vector_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("FeatureVectorizer"),
                inputdimensions=[
                    variable.type.shape[1] for variable in operator.inputs
                ],
            )
            container.add_node(
                op_type,
                feature_vector_name,
                operator.output_full_names,
                op_domain="ai.onnx.ml",
                **attrs
            )
        else:
            container.add_node(
                op_type,
                operator.input_full_names,
                operator.output_full_names,
                op_domain="ai.onnx.ml",
                **attrs
            )
    else:
        # For classifiers, due to the different representation of classes' probabilities, we need to add some
        # operators for type conversion. It turns out that we have the following topology.
        # input features ---> TreeEnsembleClassifier ---> label (must present)
        #                               |
        #                               '--> probability tensor ---> ZipMap ---> probability map (optional)
        #
        # Similar to the regressor's case, if there are multiple input features, we need to concatenate them all
        # together before feeding them into ONNX tree-based model. It leads to the following computational graph.
        #
        # input feature 1 -----.
        #        ...           |
        #        ...           v
        #        ...      ---> Feature Vectorizer ---> TreeEnsembleClassifier ---> label (must present)
        #        ...           ^                                 |
        #        ...           |                                 '--> probability tensor ---> ZipMap ---> probability
        # input feature N -----'                                                                          map (optional)

        # Set up input feature(s)
        if len(operator.inputs) > 1:
            feature_vector_name = scope.get_unique_variable_name("feature_vector")
            container.add_node(
                "FeatureVectorizer",
                operator.input_full_names,
                feature_vector_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("FeatureVectorizer"),
                inputdimensions=[
                    variable.type.shape[1] for variable in operator.inputs
                ],
            )
        else:
            feature_vector_name = operator.inputs[0].full_name

        # Find label name and probability name
        proba_output_name = None
        for variable in operator.outputs:
            if raw_model.description.predictedFeatureName == variable.raw_name:
                label_output_name = variable.full_name
            if (
                raw_model.description.predictedProbabilitiesName != ""
                and raw_model.description.predictedProbabilitiesName
                == variable.raw_name
            ):
                proba_output_name = variable.full_name

        proba_tensor_name = scope.get_unique_variable_name("ProbabilityTensor")

        if proba_output_name is not None:
            # Add tree model ONNX node with probability output
            container.add_node(
                op_type,
                feature_vector_name,
                [label_output_name, proba_tensor_name],
                op_domain="ai.onnx.ml",
                **attrs
            )

            # Add ZipMap to convert probability tensor into probability map
            container.add_node(
                "ZipMap",
                [proba_tensor_name],
                [proba_output_name],
                op_domain="ai.onnx.ml",
                **zipmap_attrs
            )
        else:
            # Add support vector classifier without probability output
            container.add_node(
                op_type,
                feature_vector_name,
                [label_output_name, proba_tensor_name],
                op_domain="ai.onnx.ml",
                **attrs
            )


register_converter("treeEnsembleClassifier", convert_tree_ensemble_model)
register_converter("treeEnsembleRegressor", convert_tree_ensemble_model)



# SPDX-License-Identifier: Apache-2.0

import json
import numpy as np
from onnx import TensorProto
from xgboost import XGBClassifier

try:
    from xgboost import XGBRFClassifier
except ImportError:
    XGBRFClassifier = None
from ...common._registration import register_converter
from ..common import get_xgb_params, get_n_estimators_classifier


class XGBConverter:
    @staticmethod
    def get_xgb_params(xgb_node):
        """
        Retrieves parameters of a model.
        """
        return get_xgb_params(xgb_node)

    @staticmethod
    def validate(xgb_node):
        params = XGBConverter.get_xgb_params(xgb_node)
        try:
            if "objective" not in params:
                raise AttributeError("ojective")
        except AttributeError as e:
            raise RuntimeError("Missing attribute in XGBoost model " + str(e))
        if hasattr(xgb_node, "missing") and not np.isnan(xgb_node.missing):
            raise RuntimeError(
                "Cannot convert a XGBoost model where missing values are not "
                "nan but {}.".format(xgb_node.missing)
            )

    @staticmethod
    def common_members(xgb_node, inputs):
        params = XGBConverter.get_xgb_params(xgb_node)
        objective = params["objective"]
        base_score = params["base_score"]
        if hasattr(xgb_node, "best_ntree_limit"):
            best_ntree_limit = xgb_node.best_ntree_limit
        elif hasattr(xgb_node, "best_iteration"):
            best_ntree_limit = xgb_node.best_iteration + 1
        else:
            best_ntree_limit = params.get("best_ntree_limit", None)
        if base_score is None:
            base_score = 0.5
        booster = xgb_node.get_booster()
        # The json format was available in October 2017.
        # XGBoost 0.7 was the first version released with it.
        js_tree_list = booster.get_dump(with_stats=True, dump_format="json")
        js_trees = [json.loads(s) for s in js_tree_list]
        return objective, base_score, js_trees, best_ntree_limit

    @staticmethod
    def _get_default_tree_attribute_pairs(is_classifier):
        attrs = {}
        for k in {
            "nodes_treeids",
            "nodes_nodeids",
            "nodes_featureids",
            "nodes_modes",
            "nodes_values",
            "nodes_truenodeids",
            "nodes_falsenodeids",
            "nodes_missing_value_tracks_true",
        }:
            attrs[k] = []
        if is_classifier:
            for k in {"class_treeids", "class_nodeids", "class_ids", "class_weights"}:
                attrs[k] = []
        else:
            for k in {
                "target_treeids",
                "target_nodeids",
                "target_ids",
                "target_weights",
            }:
                attrs[k] = []
        return attrs

    @staticmethod
    def _add_node(
        attr_pairs,
        is_classifier,
        tree_id,
        tree_weight,
        node_id,
        feature_id,
        mode,
        value,
        true_child_id,
        false_child_id,
        weights,
        weight_id_bias,
        missing,
        hitrate,
    ):
        if isinstance(feature_id, str):
            # Something like f0, f1...
            if feature_id[0] == "f":
                try:
                    feature_id = int(feature_id[1:])
                except ValueError:
                    raise RuntimeError(
                        "Unable to interpret '{0}', feature "
                        "names should follow pattern 'f%d'.".format(feature_id)
                    )
            else:
                try:
                    feature_id = int(float(feature_id))
                except ValueError:
                    raise RuntimeError(
                        "Unable to interpret '{0}', feature "
                        "names should follow pattern 'f%d'.".format(feature_id)
                    )

        # Split condition for sklearn
        # * if X_ptr[X_sample_stride * i + X_fx_stride * node.feature] <= node.threshold:
        # * https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/tree/_tree.pyx#L946
        # Split condition for xgboost
        # * if (fvalue < split_value)
        # * https://github.com/dmlc/xgboost/blob/master/include/xgboost/tree_model.h#L804

        attr_pairs["nodes_treeids"].append(tree_id)
        attr_pairs["nodes_nodeids"].append(node_id)
        attr_pairs["nodes_featureids"].append(feature_id)
        attr_pairs["nodes_modes"].append(mode)
        attr_pairs["nodes_values"].append(float(value))
        attr_pairs["nodes_truenodeids"].append(true_child_id)
        attr_pairs["nodes_falsenodeids"].append(false_child_id)
        attr_pairs["nodes_missing_value_tracks_true"].append(missing)
        if "nodes_hitrates" in attr_pairs:
            attr_pairs["nodes_hitrates"].append(hitrate)
        if mode == "LEAF":
            if is_classifier:
                for i, w in enumerate(weights):
                    attr_pairs["class_treeids"].append(tree_id)
                    attr_pairs["class_nodeids"].append(node_id)
                    attr_pairs["class_ids"].append(i + weight_id_bias)
                    attr_pairs["class_weights"].append(float(tree_weight * w))
            else:
                for i, w in enumerate(weights):
                    attr_pairs["target_treeids"].append(tree_id)
                    attr_pairs["target_nodeids"].append(node_id)
                    attr_pairs["target_ids"].append(i + weight_id_bias)
                    attr_pairs["target_weights"].append(float(tree_weight * w))

    @staticmethod
    def _fill_node_attributes(
        treeid, tree_weight, jsnode, attr_pairs, is_classifier, remap
    ):
        if "children" in jsnode:
            XGBConverter._add_node(
                attr_pairs=attr_pairs,
                is_classifier=is_classifier,
                tree_id=treeid,
                tree_weight=tree_weight,
                value=jsnode["split_condition"],
                node_id=remap[jsnode["nodeid"]],
                feature_id=jsnode["split"],
                mode="BRANCH_LT",  # 'BRANCH_LEQ' --> is for sklearn
                true_child_id=remap[jsnode["yes"]],  # ['children'][0]['nodeid'],
                false_child_id=remap[jsnode["no"]],  # ['children'][1]['nodeid'],
                weights=None,
                weight_id_bias=None,
                missing=jsnode.get("missing", -1) == jsnode["yes"],
                hitrate=jsnode.get("cover", 0),
            )

            for ch in jsnode["children"]:
                if "children" in ch or "leaf" in ch:
                    XGBConverter._fill_node_attributes(
                        treeid, tree_weight, ch, attr_pairs, is_classifier, remap
                    )
                else:
                    raise RuntimeError("Unable to convert this node {0}".format(ch))

        else:
            weights = [jsnode["leaf"]]
            weights_id_bias = 0
            XGBConverter._add_node(
                attr_pairs=attr_pairs,
                is_classifier=is_classifier,
                tree_id=treeid,
                tree_weight=tree_weight,
                value=0.0,
                node_id=remap[jsnode["nodeid"]],
                feature_id=0,
                mode="LEAF",
                true_child_id=0,
                false_child_id=0,
                weights=weights,
                weight_id_bias=weights_id_bias,
                missing=False,
                hitrate=jsnode.get("cover", 0),
            )

    @staticmethod
    def _remap_nodeid(jsnode, remap=None):
        if remap is None:
            remap = {}
        nid = jsnode["nodeid"]
        remap[nid] = len(remap)
        if "children" in jsnode:
            for ch in jsnode["children"]:
                XGBConverter._remap_nodeid(ch, remap)
        return remap

    @staticmethod
    def fill_tree_attributes(js_xgb_node, attr_pairs, tree_weights, is_classifier):
        if not isinstance(js_xgb_node, list):
            raise TypeError("js_xgb_node must be a list")
        for treeid, (jstree, w) in enumerate(zip(js_xgb_node, tree_weights)):
            remap = XGBConverter._remap_nodeid(jstree)
            XGBConverter._fill_node_attributes(
                treeid, w, jstree, attr_pairs, is_classifier, remap
            )


class XGBRegressorConverter(XGBConverter):
    @staticmethod
    def validate(xgb_node):
        return XGBConverter.validate(xgb_node)

    @staticmethod
    def _get_default_tree_attribute_pairs():
        attrs = XGBConverter._get_default_tree_attribute_pairs(False)
        attrs["post_transform"] = "NONE"
        attrs["n_targets"] = 1
        return attrs

    @staticmethod
    def convert(scope, operator, container):
        xgb_node = operator.raw_operator
        inputs = operator.inputs
        objective, base_score, js_trees, best_ntree_limit = XGBConverter.common_members(
            xgb_node, inputs
        )

        if objective in ["reg:gamma", "reg:tweedie"]:
            raise RuntimeError("Objective '{}' not supported.".format(objective))

        attr_pairs = XGBRegressorConverter._get_default_tree_attribute_pairs()
        attr_pairs["base_values"] = [base_score]

        if best_ntree_limit and best_ntree_limit < len(js_trees):
            js_trees = js_trees[:best_ntree_limit]

        XGBConverter.fill_tree_attributes(
            js_trees, attr_pairs, [1 for _ in js_trees], False
        )

        # add nodes
        if objective == "count:poisson":
            names = [scope.get_unique_variable_name("tree")]
            del attr_pairs["base_values"]
        else:
            names = operator.output_full_names
        container.add_node(
            "TreeEnsembleRegressor",
            operator.input_full_names,
            names,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("TreeEnsembleRegressor"),
            **attr_pairs,
        )

        if objective == "count:poisson":
            cst = scope.get_unique_variable_name("poisson")
            container.add_initializer(cst, TensorProto.FLOAT, [1], [base_score])
            new_name = scope.get_unique_variable_name("exp")
            container.add_node("Exp", names, [new_name])
            container.add_node("Mul", [new_name, cst], operator.output_full_names)


class XGBClassifierConverter(XGBConverter):
    @staticmethod
    def validate(xgb_node):
        return XGBConverter.validate(xgb_node)

    @staticmethod
    def _get_default_tree_attribute_pairs():
        attrs = XGBConverter._get_default_tree_attribute_pairs(True)
        # TODO: check it is implemented. The model cannot
        # be loaded when they are present.
        # attrs['nodes_hitrates'] = []
        return attrs

    @staticmethod
    def convert(scope, operator, container):
        xgb_node = operator.raw_operator
        inputs = operator.inputs

        objective, base_score, js_trees, best_ntree_limit = XGBConverter.common_members(
            xgb_node, inputs
        )

        params = XGBConverter.get_xgb_params(xgb_node)
        n_estimators = get_n_estimators_classifier(xgb_node, params, js_trees)
        num_class = params.get("num_class", None)

        attr_pairs = XGBClassifierConverter._get_default_tree_attribute_pairs()
        XGBConverter.fill_tree_attributes(
            js_trees, attr_pairs, [1 for _ in js_trees], True
        )
        if num_class is not None:
            ncl = num_class
            n_estimators = len(js_trees) // ncl
        else:
            ncl = (max(attr_pairs["class_treeids"]) + 1) // n_estimators

        best_ntree_limit = best_ntree_limit or len(js_trees)
        if ncl > 0:
            best_ntree_limit *= ncl
        if 0 < best_ntree_limit < len(js_trees):
            js_trees = js_trees[:best_ntree_limit]
            attr_pairs = XGBClassifierConverter._get_default_tree_attribute_pairs()
            XGBConverter.fill_tree_attributes(
                js_trees, attr_pairs, [1 for _ in js_trees], True
            )

        if len(attr_pairs["class_treeids"]) == 0:
            raise RuntimeError("XGBoost model is empty.")

        if ncl <= 1:
            ncl = 2
            if objective != "binary:hinge":
                # See https://github.com/dmlc/xgboost/blob/master/src/common/math.h#L23.
                attr_pairs["post_transform"] = "LOGISTIC"
                attr_pairs["class_ids"] = [0 for v in attr_pairs["class_treeids"]]
                if js_trees[0].get("leaf", None) == 0:
                    attr_pairs["base_values"] = [base_score]
                elif base_score != 0.5:
                    # 0.5 -> cst = 0
                    cst = -np.log(1 / np.float32(base_score) - 1.0)
                    attr_pairs["base_values"] = [cst]
            else:
                attr_pairs["base_values"] = [base_score]
        else:
            # See https://github.com/dmlc/xgboost/blob/master/src/common/math.h#L35.
            attr_pairs["post_transform"] = "SOFTMAX"
            attr_pairs["base_values"] = [base_score for n in range(ncl)]
            attr_pairs["class_ids"] = [v % ncl for v in attr_pairs["class_treeids"]]

        classes = xgb_node.classes_
        if (
            np.issubdtype(classes.dtype, np.floating)
            or np.issubdtype(classes.dtype, np.integer)
            or np.issubdtype(classes.dtype, np.bool_)
        ):
            attr_pairs["classlabels_int64s"] = classes.astype("int")
        else:
            classes = np.array([s.encode("utf-8") for s in classes])
            attr_pairs["classlabels_strings"] = classes

        # add nodes
        if objective in ("binary:logistic", "binary:hinge"):
            ncl = 2
            if objective == "binary:hinge":
                attr_pairs["post_transform"] = "NONE"
                output_names = [
                    operator.output_full_names[0],
                    scope.get_unique_variable_name("output_prob"),
                ]
            else:
                output_names = operator.output_full_names
            container.add_node(
                "TreeEnsembleClassifier",
                operator.input_full_names,
                output_names,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("TreeEnsembleClassifier"),
                **attr_pairs,
            )
            if objective == "binary:hinge":
                if container.target_opset < 9:
                    raise RuntimeError(
                        f"hinge function cannot be implemented because "
                        f"opset={container.target_opset}<9."
                    )
                zero = scope.get_unique_variable_name("zero")
                one = scope.get_unique_variable_name("one")
                container.add_initializer(zero, TensorProto.FLOAT, [1], [0.0])
                container.add_initializer(one, TensorProto.FLOAT, [1], [1.0])
                greater = scope.get_unique_variable_name("output_prob")
                container.add_node("Greater", [output_names[1], zero], [greater])
                container.add_node(
                    "Where", [greater, one, zero], operator.output_full_names[1]
                )
        elif objective in ("multi:softprob", "multi:softmax"):
            ncl = len(js_trees) // n_estimators
            if objective == "multi:softmax":
                attr_pairs["post_transform"] = "NONE"
            container.add_node(
                "TreeEnsembleClassifier",
                operator.input_full_names,
                operator.output_full_names,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("TreeEnsembleClassifier"),
                **attr_pairs,
            )
        elif objective == "reg:logistic":
            ncl = len(js_trees) // n_estimators
            if ncl == 1:
                ncl = 2
            container.add_node(
                "TreeEnsembleClassifier",
                operator.input_full_names,
                operator.output_full_names,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("TreeEnsembleClassifier"),
                **attr_pairs,
            )
        else:
            raise RuntimeError("Unexpected objective: {0}".format(objective))


def convert_xgboost(scope, operator, container):
    xgb_node = operator.raw_operator
    if isinstance(xgb_node, (XGBClassifier, XGBRFClassifier)) or getattr(
        xgb_node, "operator_name", None
    ) in ("XGBClassifier", "XGBRFClassifier"):
        cls = XGBClassifierConverter
    else:
        cls = XGBRegressorConverter
    cls.validate(xgb_node)
    cls.convert(scope, operator, container)


register_converter("XGBClassifier", convert_xgboost)
register_converter("XGBRFClassifier", convert_xgboost)
register_converter("XGBRegressor", convert_xgboost)
register_converter("XGBRFRegressor", convert_xgboost)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_slice_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C', H, W]
        2. [N, C, H, W] ---> [N, C, H', W]
        3. [N, C, H, W] ---> [N, C, H, W']
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)

    params = operator.raw_operator.slice

    from coremltools.proto.NeuralNetwork_pb2 import SliceLayerParams as Params

    axis_map = {Params.CHANNEL_AXIS: 1, Params.HEIGHT_AXIS: 2, Params.WIDTH_AXIS: 3}

    if params.startIndex >= 0:
        output_shape[axis_map[Params.CHANNEL_AXIS]] = (
            params.endIndex - params.startIndex
        )
    else:
        output_shape[axis_map[Params.CHANNEL_AXIS]] += (
            1 + params.endIndex - params.startIndex
        )

    operator.outputs[0].type = FloatTensorType(
        output_shape, doc_string=operator.outputs[0].type.doc_string
    )


register_shape_calculator("slice", calculate_slice_output_shapes)



# SPDX-License-Identifier: Apache-2.0

# To register shape calculators for Core ML neural
# network operators, import associated modules here.
from . import BatchNorm
from . import BidirectionalLSTM
from . import Concat
from . import Convolution
from . import Crop
from . import Dot
from . import Embed
from . import Flatten
from . import GRU
from . import IdentityFloat
from . import InnerProduct
from . import LoadConstant
from . import LoadConstantND
from . import LSTM
from . import Merge
from . import Pad
from . import Permute
from . import Pool
from . import Reduce
from . import ReorganizeData
from . import Reshape
from . import ReshapeStatic
from . import SequenceRepeat
from . import Slice
from . import Split
from . import Upsample



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_reshape
from ....common._registration import register_converter


def convert_inner_product(scope, operator, container):
    params = operator.raw_operator.innerProduct

    # Apply pre-processing step if needed
    if len(operator.inputs[0].type.shape) == 4:
        # Input shape is [N, C, 1, 1]. Adjust input shape
        # because Gemm in ONNX only takes 2-D input
        reshaped_tensor_name = scope.get_unique_variable_name(
            operator.inputs[0].full_name + "_reshaped"
        )
        apply_reshape(
            scope,
            operator.inputs[0].full_name,
            reshaped_tensor_name,
            container,
            desired_shape=[-1, int(params.inputChannels)],
        )
        name_a = reshaped_tensor_name
    else:
        # Input shape is [N, C]. There is no pre-processing for applying ONNX operator.
        name_a = operator.inputs[0].full_name

    # Allocate the weights of dense layer
    name_b = scope.get_unique_variable_name(operator.full_name + "_B")
    shape_b = [params.outputChannels, params.inputChannels]
    container.add_initializer(
        name_b, onnx_proto.TensorProto.FLOAT, shape_b, params.weights.floatValue
    )

    # Allocate the bias of dense layer
    name_c = scope.get_unique_variable_name(operator.full_name + "_C")
    shape_c = [params.outputChannels]
    if params.hasBias:
        container.add_initializer(
            name_c, onnx_proto.TensorProto.FLOAT, shape_c, params.bias.floatValue
        )
    else:
        container.add_initializer(
            name_c, onnx_proto.TensorProto.FLOAT, shape_c, [0.0] * shape_b[0]
        )

    # Set up attributes for ONNX Gemm which is the counterpart
    # of CoreML inner product layer in ONNX.
    attrs = {"name": operator.full_name}
    attrs["alpha"] = 1.0
    attrs["beta"] = 1.0
    attrs["transA"] = 0
    attrs["transB"] = 1

    # Get the correct version number for Gemm in ONNX
    if container.target_opset < 5:
        attrs["broadcast"] = 1
        op_version = 1
    elif container.target_opset < 7:
        attrs["broadcast"] = 1
        op_version = 6
    elif container.target_opset < 9:
        op_version = 7
    elif container.target_opset < 11:
        op_version = 9
    else:
        op_version = 11

    # Create the major ONNX operator, Gemm, to do CoreML
    # inner product and possibly add shape adjustment
    if len(operator.inputs[0].type.shape) == 4:
        # Input shape is [N, C, 1, 1] so we expect output is also 4-D, [N, C', 1, 1].
        buffer_tensor_name = scope.get_unique_variable_name(
            operator.full_name + "_buffer"
        )
        container.add_node(
            "Gemm",
            [name_a, name_b, name_c],
            buffer_tensor_name,
            op_version=op_version,
            **attrs
        )
        apply_reshape(
            scope,
            buffer_tensor_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[-1, int(params.outputChannels), 1, 1],
        )
    else:
        # Input shape is [N, C], so we don't need to change Gemm's output shape.
        container.add_node(
            "Gemm",
            [name_a, name_b, name_c],
            operator.outputs[0].full_name,
            op_version=op_version,
            **attrs
        )


register_converter("innerProduct", convert_inner_product)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
from pyspark.ml.feature import Imputer
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


## For some reason during the spark bring up and shutdown
## something happens causing Imputer
## tests to fail. For that you need to run each test here individually
## for now these will be commented out so as not to break the build
##      AttributeError: 'NoneType' object has no attribute
# 'setCallSite' on model.surrogateDF
## Therefore we leave these tests out for now until a newere
## version of pyspark is availabe that address this issue
class TestSparkmlImputer(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_imputer_single(self):
        self._imputer_test_single()

    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_imputer_multi(self):
        self._imputer_test_multi()

    def _imputer_test_multi(self):
        data = self.spark.createDataFrame(
            [
                (1.0, float("nan")),
                (2.0, float("nan")),
                (float("nan"), 3.0),
                (4.0, 4.0),
                (5.0, 5.0),
            ],
            ["a", "b"],
        )
        imputer = Imputer(inputCols=["a", "b"], outputCols=["out_a", "out_b"])
        model = imputer.fit(data)

        # the input name should match the inputCols above
        model_onnx = convert_sparkml(
            model,
            "Sparkml Imputer Multi Input",
            [("a", FloatTensorType([None, 1])), ("b", FloatTensorType([None, 1]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)

        expected = {
            "out_a": predicted.select("out_a").toPandas().values.astype(numpy.int64),
            "out_b": predicted.select("out_b").toPandas().values.astype(numpy.int64),
        }

        data_np = data.toPandas().values.astype(numpy.float32)
        data_np = {"a": data_np[:, :1], "b": data_np[:, 1:]}
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlImputerMulti"
        )
        onnx_model_path = paths[-1]
        output_names = ["out_a", "out_b"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))
        compare_results(expected, actual_output, decimal=5)

    def _imputer_test_single(self):
        data = self.spark.createDataFrame(
            [
                (1.0, float("nan")),
                (2.0, float("nan")),
                (float("nan"), 3.0),
                (4.0, 4.0),
                (5.0, 5.0),
            ],
            ["a", "b"],
        )
        imputer = Imputer(inputCols=["a"], outputCols=["out_a"])
        model = imputer.fit(data)

        # the input name should match the inputCols above
        model_onnx = convert_sparkml(
            model,
            "Sparkml Imputer",
            [("a", FloatTensorType([None, 1]))],
            target_opset=TARGET_OPSET,
        )
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = predicted.select("out_a").toPandas().values.astype(numpy.float32)
        data_np = data.toPandas().a.values.astype(numpy.float32)
        data_np = data_np.reshape((-1, 1))
        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlImputerSingle"
        )
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(["out_a"], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import numpy as np

from ....common._apply_operation import (
    apply_elu,
    apply_hard_sigmoid,
    apply_leaky_relu,
    apply_prelu,
    apply_relu,
    apply_sigmoid,
    apply_tanh,
    apply_affine,
    apply_parametric_softplus,
    apply_scaled_tanh,
    apply_thresholded_relu,
)
from ....common._registration import register_converter


def convert_activation(scope, operator, container):
    inputs = [variable.full_name for variable in operator.inputs]
    outputs = [variable.full_name for variable in operator.outputs]
    attrs = {"name": operator.full_name}

    params = operator.raw_operator.activation
    activation_type = params.WhichOneof("NonlinearityType")

    # Create ONNX objects by high-level APIs such as apply_relu(...) if possible.
    # Otherwise, we use low-level APIs such as add_node(...)
    if activation_type == "leakyReLU":
        apply_leaky_relu(
            scope,
            inputs,
            outputs,
            container,
            operator_name=attrs["name"],
            alpha=params.leakyReLU.alpha,
        )
    elif activation_type == "ReLU":
        apply_relu(scope, inputs, outputs, container, operator_name=attrs["name"])
    elif activation_type == "PReLU":
        apply_prelu(
            scope,
            inputs[0],
            outputs,
            container,
            operator_name=attrs["name"],
            slope=np.asarray([params.PReLU.alpha.floatValue]),
        )
    elif activation_type == "ELU":
        apply_elu(
            scope,
            inputs,
            outputs,
            container,
            operator_name=attrs["name"],
            alpha=params.ELU.alpha,
        )
    elif activation_type == "tanh":
        apply_tanh(scope, inputs, outputs, container, operator_name=attrs["name"])
    elif activation_type == "sigmoid":
        apply_sigmoid(scope, inputs, outputs, container, operator_name=attrs["name"])
    elif activation_type == "sigmoidHard":
        apply_hard_sigmoid(
            scope,
            inputs,
            outputs,
            container,
            operator_name=attrs["name"],
            alpha=params.sigmoidHard.alpha,
            beta=params.sigmoidHard.beta,
        )
    elif activation_type == "linear":
        apply_affine(
            scope,
            inputs[0],
            outputs[0],
            container,
            operator_name=attrs["name"],
            alpha=params.linear.alpha,
            beta=params.linear.beta,
        )
    elif activation_type == "parametricSoftplus":
        apply_parametric_softplus(
            scope,
            inputs[0],
            outputs[0],
            container,
            operator_name=attrs["name"],
            alpha=params.parametricSoftplus.alpha.floatValue,
            beta=params.parametricSoftplus.beta.floatValue,
        )
    elif activation_type == "scaledTanh":
        apply_scaled_tanh(
            scope,
            inputs[0],
            outputs[0],
            container,
            operator_name=attrs["name"],
            alpha=params.scaledTanh.alpha,
            beta=params.scaledTanh.beta,
        )
    elif activation_type == "thresholdedReLU":
        apply_thresholded_relu(
            scope,
            inputs,
            outputs,
            container,
            operator_name=attrs["name"],
            alpha=params.thresholdedReLU.alpha,
        )
    else:
        if activation_type == "softsign":
            op_type = "Softsign"
        elif activation_type == "softplus":
            op_type = "Softplus"
        else:
            raise TypeError("Unsupported activation layer {0}".format(activation_type))

        container.add_node(op_type, inputs, outputs, **attrs)


register_converter("activation", convert_activation)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter
from .SVC import extract_support_vectors_as_dense_tensor


def convert_svm_regressor(scope, operator, container):
    params = operator.raw_operator.supportVectorRegressor

    kernel_enum = {
        "linearKernel": "LINEAR",
        "polyKernel": "POLY",
        "rbfKernel": "RBF",
        "sigmoidKernel": "SIGMOID",
        "precomputedKernel": "PRECOMPUTED",
    }
    kernel = params.kernel
    kernel_val = kernel.WhichOneof("kernel")
    svr_kernel = kernel_enum[kernel_val]

    if kernel_val == "rbfKernel":
        svr_kernel_params = [kernel.rbfKernel.gamma, 0.0, 0.0]
    elif kernel_val == "polyKernel":
        svr_kernel_params = [
            kernel.polyKernel.gamma,
            kernel.polyKernel.coef0,
            kernel.polyKernel.degree,
        ]
    elif kernel_val == "sigmoidKernel":
        svr_kernel_params = [
            kernel.sigmoidKernel.gamma,
            kernel.sigmoidKernel.coef0,
            0.0,
        ]
    elif kernel_val == "linearKernel":
        svr_kernel_params = [0.0, 0.0, 0.0]

    n_supports, support_vectors = extract_support_vectors_as_dense_tensor(
        operator.raw_operator.supportVectorRegressor
    )

    svr_coefficients = params.coefficients.alpha
    if isinstance(params.rho, list):
        svr_rho = [-x for x in params.rho]
    else:
        svr_rho = [-params.rho]

    op_type = "SVMRegressor"
    op_name = scope.get_unique_operator_name(op_type)
    attrs = {"name": op_name}
    attrs["kernel_type"] = svr_kernel
    attrs["kernel_params"] = svr_kernel_params
    attrs["support_vectors"] = support_vectors
    attrs["n_supports"] = n_supports
    attrs["coefficients"] = svr_coefficients
    attrs["rho"] = svr_rho

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("supportVectorRegressor", convert_svm_regressor)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import FloatTensorType, Int64TensorType, FloatType, Int64Type
from ...common.utils import check_input_and_output_types


def calculate_traditional_regressor_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, C']

    The number C' is the length of prediction vector.
    It can be a scalar (C'=1) or a vector (C'>1)
    """
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, Int64TensorType, FloatType, Int64Type],
    )

    if any(len(variable.type.shape) != 2 for variable in operator.inputs):
        raise RuntimeError("Input(s) must be 2-D tensor(s)")

    model_type = operator.raw_operator.WhichOneof("Type")
    if model_type == "glmRegressor":
        glm = operator.raw_operator.glmRegressor
        C = len(glm.weights)
    elif model_type == "treeEnsembleRegressor":
        tree = operator.raw_operator.treeEnsembleRegressor.treeEnsemble
        C = len(tree.basePredictionValue)
    elif model_type == "supportVectorRegressor":
        C = 1
    else:
        raise ValueError(
            "Model should be one of linear model, tree-based model, and support vector machine"
        )

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType(
        [N, C], doc_string=operator.outputs[0].type.doc_string
    )


register_shape_calculator("glmRegressor", calculate_traditional_regressor_output_shapes)
register_shape_calculator(
    "supportVectorRegressor", calculate_traditional_regressor_output_shapes
)
register_shape_calculator(
    "treeEnsembleRegressor", calculate_traditional_regressor_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import (
    DictionaryType,
    FloatTensorType,
    SequenceType,
    StringTensorType,
    Int64TensorType,
)
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_tensor_to_probability_map_output_shapes(operator):
    """
    Allowed input/output patterns are
    ONNX < 1.2
        1. [1, C] ---> ---> A map
        2. [1, C_1, ..., C_n] ---> A map
    ONNX >= 1.2
        1. [N, C] ---> ---> A sequence of maps
        2. [N, C_1, ..., C_n] ---> A sequence of maps

    Note that N must be 1 currently if you're using
    ONNX<1.2 because old ZipMap doesn't produce a seqneuce of map If the
    input is not [N, C], it will be reshaped into
    [N, C_1 x C_2, x ... x C_n] before being fed into ONNX ZipMap.
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    model_type = operator.raw_operator.WhichOneof("Type")
    if model_type == "neuralNetworkClassifier":
        class_label_type = operator.raw_operator.neuralNetworkClassifier.WhichOneof(
            "ClassLabels"
        )
    else:
        raise TypeError("%s has no class label" % model_type)

    N = operator.inputs[0].type.shape[0]
    doc_string = operator.outputs[0].type.doc_string
    if class_label_type == "stringClassLabels":
        if operator.target_opset < 7:
            operator.outputs[0].type = DictionaryType(
                StringTensorType([1]), FloatTensorType([1]), doc_string
            )
        else:
            operator.outputs[0].type = SequenceType(
                DictionaryType(StringTensorType([]), FloatTensorType([])), N, doc_string
            )
    elif class_label_type == "int64ClassLabels":
        if operator.target_opset < 7:
            operator.outputs[0].type = DictionaryType(
                Int64TensorType([1]), FloatTensorType([1]), doc_string
            )
        else:
            operator.outputs[0].type = SequenceType(
                DictionaryType(Int64TensorType([]), FloatTensorType([])), N, doc_string
            )
    else:
        raise ValueError("Unsupported label type")


register_shape_calculator(
    "tensorToProbabilityMap", calculate_tensor_to_probability_map_output_shapes
)
